# -*- encoding:utf-8 -*-
"""
    主裁基础实现模块
"""
import os, pdb
from os import path
from pyexpat import model
import numpy as np
import pandas as pd
import math, copy
from abc import abstractmethod
import scipy.optimize as sco
from ultron.ump.core import env as UMPEnv
from ultron.kdutils.parallel import delayed, Parallel
from ultron.kdutils import file as file_util
from ultron.ump.core.fixes import GMM
from ultron.kdutils.progress import Progress
from ultron.ump.ump.base import Base
"""主裁使用的gmm默认分类从40-85个分类"""
K_DEFAULT_NCS_RANG = slice(40, 85, 1)
"""主裁使用的gmm默认分类K_DEFAULT_NCS_RANG最大对应的分类数量，详阅读使用"""
K_DEFAULT_NCS_MAX = 50000
"""在brust_min中计算lps_step, lms_step, lrs_step中使用的默认step参数，默认50，详见brust_min函数实现"""
g_brust_min_step_cnt = 50
"""在不计算全局最优参数brust_min组合情况下的直接使用的默认参数摘取cprs形成llps"""
g_brust_min_default = (0, 0, 0.65)
"""代表在ump_main_clf_dump中show_order或者save_order为True的情况下最多绘制和保存的交易快照数量"""
g_plot_order_max_cnt = 100

g_cpu_cnt = 4  ## fix me

root_drive = path.expanduser('~')  ## fix me

g_project_root = path.join(root_drive, 'ultron')  ## fix me

g_project_data_dir = path.join(g_project_root, 'data')  ## fix me

g_project_cache_dir = path.join(g_project_data_dir, 'cache')  ## fix me


def _do_gmm_cluster(sub_ncs, x, df, threshold):
    """
    在MainBase中fit并行启动的子进程执行的gmm cluster函数，进程函数，
    在子进程中迭代sub_ncs中的component值代入gmm，通过threshold对gmm cluster
    结果进行筛选过滤
    :param sub_ncs: 子进程中gmm分类的范围, eg： [10, 11, 12, 13, 14, 15]
    :param x: 主裁训练集特征x矩阵，numpy矩阵对象
    :param df: 主裁训练集特征pd.DataFrame对象, 包括x，y
    :param threshold:  分类簇中失败率选择的阀值（默认0.65），即大于threshold值的gmm分类簇做为主裁学习目标分类簇
    :return: clf_component_dict, cluster_df_dict
    """
    clf_component_dict = {}
    cluster_df_dict = {}
    # 启动多进程进度显示UltronMulPidProgress
    with Progress(len(sub_ncs), 'gmm fit') as progress:
        for epoch, component in enumerate(sub_ncs):
            clf = GMM(component, random_state=3).fit(x)
            cluster = clf.predict(x)
            """
                eg：component=14, cluster形式如：

                      [ 4,  4, 13,  2,  3, 13,  3,  7,  1,  8,  4, 13,  4,  4,  2, 13, 13,
                        ....... 1,  1,  5,  7,  5,  7,  0,  8, 13, 10, 10,  2,  2,  7, 12,
                      12, 13,  7,  7, 13, 13]
            """
            # 只是临时保存一下分类簇序号，因为要使用pd.crosstab，下一个迭代中新的cluster又会生成
            df['cluster'] = cluster
            xt = pd.crosstab(df['cluster'], df['result'])
            """
                xt形如: 即gmm分类簇中每一个子分类的失败交易和盈利交易的数量统计
                result       0      1
                cluster
                0          290    279
                2         1156    766
                3          160    137
                .....................
            """
            # 进行一次cluster数量的淘汰，无法准确量化某一个分类数量少于多少为不正常，这里范定义总交易数的1/1000
            xt = xt[xt.sum(axis=1) > (df.shape[0] / 1000)]

            xt_pct = xt.div(xt.sum(1).astype(float), axis=0)
            """
                xt_pct形如: 即gmm分类簇中每一个子分类的失败交易和盈利交易的数量占比例
                            目的是方便和参数中的threshold（默认0.65）进行阀值对比
                result          0         1
                cluster
                0        0.509666  0.490334
                2        0.601457  0.398543
                3        0.538721  0.461279
                ............................
            """
            # xt_pct[0]即为分类簇中每一个子分类的失败率
            cluster_ind = xt_pct[xt_pct[0] > threshold].index
            """
                eg：xt_pct[xt_pct[0] > threshold].index的返回结果rt形式如下：
                Int64Index([7, 8, 9, 10], dtype='int64', name='cluster')

                cluster_ind的类型为pd.Index序列对象，值序列[7, 8, 9, 10]即代表使用GMM(component)分类
                下分成component个分类中，满足子分类的失败率大于阀值threshold的子分类序号
            """
            if len(cluster_ind) > 0:
                # 把失败概率大于阀值的component的保存clf以及cluster_ind
                clf_component_dict[component] = (clf, cluster_ind)
                """
                    eg：clf_component_dict字典形式如下：
                        key=component, value=(GaussianMixtured对象，cluster_ind: GMM(component)分component个类中，
                                              满足子分类的失败率大于阀值threshold的子分类序号)

                    {14: (GaussianMixture(max_iter=100, n_components=14, n_init=1),
                        Int64Index([7, 8, 9, 10], dtype='int64', name='cluster')),

                    15: (GaussianMixture(max_iter=100, n_components=15, n_init=1),
                        Int64Index([7, 8, 9, 10], dtype='int64', name='cluster'))}
                """
                # component下的大于阀值的子分类cluster_ind进行迭代
                for cluster in cluster_ind:
                    # cluster_df_key = component + cluster, eg: '14_7'
                    cluster_df_key = '{0}_{1}'.format(component, cluster)
                    # 从df中取出cluster对于的子pd.DataFrame对象cluster_df
                    cluster_df = df[df['cluster'] == cluster]
                    """
                        eg: cluster_df
                                    result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
                        2014-11-11       1          8.341          -9.450          0.730
                        2015-10-28       0          7.144          -9.818         -3.886
                        2015-11-04       0         12.442         -10.353          3.313
                        2016-03-30       0         13.121          -8.461          4.498
                        2016-04-15       0          4.238         -13.247          4.693
                        2016-04-15       0          4.238         -13.247          4.693

                                    buy_deg_ang21  ind  cluster
                        2014-11-11         12.397    7        7
                        2015-10-28          6.955   39        7
                        2015-11-04          7.840   41        7
                        2016-03-30          4.070   49        7
                        2016-04-15          1.162   53        7
                        2016-04-15          1.162   54        7
                    """
                    # 以cluster_df_key做为key， value=cluster_df保存在cluster_df_dict中
                    cluster_df_dict[cluster_df_key] = cluster_df
    return clf_component_dict, cluster_df_dict


class MainBase(Base):
    """主裁基类"""

    @classmethod
    def ump_main_clf_dump(cls,
                          orders_pd_train,
                          p_ncs=None,
                          brust_min=True,
                          market_name=None,
                          show_info=False):
        """
        类方法，通过交易训练集orders_pd_train构造UltronUmpMainBase子类对象，透传brust_min，p_ncs等参数
        使用fit方法对训练集进行分类簇筛选，使用dump_clf本地序列化训练结果
        :param orders_pd_train: 交易训练集，pd.DataFrame对象
        :param p_ncs: gmm分类的范围, 可以为具体序列对象如[10, 11, 12....80], 也可以为生成器对象，
                      如xrange(10, 80, 10)，还支持slice对象，eg：slice(10, 80, 10)
        :param brust_min: bool类型，代表是否进行全局最优参数计算brust_min，如果否则直接使用默认g_brust_min_default(0, 0, 0.65)
                          如果不使用brust_min，即brust_min=False可大大提高训练运行效率
        :param market_name: 主裁训练或者获取裁判对应的存贮唯一名称，默认None, 根据env中的当前市场设置存储名称
        :param show_component: 是否可视化lcs，lrs，lps，lms数据(2d, 3d)，即透传给fit函数的show参数
        :param show_info: 是否显示brust_min计算的最优lrs，lps，lms组合结果以及best_hit_cnt_info，choose_cprs_component
                          等辅助函数的辅助输出信息
        :param save_order: 是否保存失败概率最大的分类簇的交易快照图片到本地
        :param show_order: 是否绘制失败概率最大的分类簇的交易快照图片
        :return: UltronUmpMainBase子类对象实例
        """
        ump = cls(orders_pd_train, market_name=market_name)
        # 训练抽取分类簇，根据brust_min
        ump.fit(p_ncs=p_ncs, brust_min=brust_min)
        ump.dump_clf()

        if show_info:
            ump.log_func('全局最优:{}'.format(ump.llps_brust_min))
            # 统计拦截hit best参数
            ump.best_hit_cnt_info(ump.llps)
            # 统计滤去llps的提升情况
            ump.choose_cprs_component(ump.llps)
        return ump

    @abstractmethod
    def get_fiter_class(self):
        """abstractmethod子类必须实现，声明具体子类裁判使用的筛选特征形成特征的类"""
        pass

    @abstractmethod
    def get_predict_col(self):
        """abstractmethod子类必须实现，获取具体子类裁判需要的特征keys"""
        pass

    @classmethod
    @abstractmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法，abstractmethod子类必须实现
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        pass

    def __init__(self,
                 orders_pd=None,
                 predict=False,
                 market_name=None,
                 **kwarg):
        """
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象, 最好是经过度量类
                          UltronMetricsBase对象进行度量fit_metrics之后的orders_pd
        :param predict: 是否构造的裁判类型为预测，非训练裁判
        :param market_name: 主裁训练或者获取裁判对应的存贮唯一名称，默认None, 根据env中的当前市场设置存储名称
        :param kwarg: 将kwarg参数透传给fiter_cls的构造：
                        self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        """
        self.orders_pd = orders_pd
        # 特征筛选类fiter_cls
        self.fiter_cls = self.get_fiter_class()
        self.log_func = print  #logging.info if UltronEnv.g_is_ipython else print
        if isinstance(market_name, UMPEnv.EMarketTargetType):
            market_name = market_name.value

        # predict或者训练的情况都需要对应裁判的唯一名称, 默认使用对应市场的字符串名字 eg，'us'， 'cn'
        self.market_name = UMPEnv.g_market_target.value if market_name is None else market_name

        if predict:
            # 如果是predict非训练目的，这里直接返回
            # TODO 拆开predict和训练数据逻辑，不要纠缠在一起
            return

        if orders_pd is not None and 'profit_cg' not in orders_pd.columns:
            # profit_cg等度量参数是要在UltronMetricsBase结束后才会有
            self.log_func(
                'you do better UltronMetricsBase.fit_metrics in orders_pd!!!!')
            from ultron.ump.metrics import base as metrics_base
            # 这里只做fit_metrics_order，没做fit_metrics因为比如期货，比特币会有自己的度量类，使用通用的fit_metrics_order
            metrics_base(orders_pd, None, None, None).fit_metrics_order()
        # 实例化特征构造对象self.fiter
        self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        """
            self.fiter是UltronMLPd子类对象，在init中即通过make_xy筛选出orders_pd中需要的训练集特征，在子类
            实现的make_xy函数都被主裁装饰器函数ump_main_make_xy装饰对筛选出的训练集特征进行转换矩阵，提取x,
            y序列等统一操作，详阅UltronUmpMainDeg等示例主裁具体子类

            这里构造好的self.fiter中df对象已经存在特定的特征
            eg: self.fiter.df
                        result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21
            2014-09-24	0	    3.378	        3.458	        3.458	        1.818
            2014-10-24	0	    0.191	        2.889	        2.809	        -1.089
            2014-10-29	1	    -2.026	        16.689      	-0.761	        1.980
            2014-10-29	1	    -3.427      	-11.956	        -8.296	        6.507
            2014-10-29	1	    -2.915	        39.469      	-6.043	        7.046
        """
        # 默认使用svm，这里需要参数可设置
        self.fiter().estimator.svc()

    # noinspection PyMethodMayBeStatic
    def _sub_ncs_split(self, ncs, n_jobs):
        sub_ncs_cnt = int(len(ncs) / n_jobs)
        if sub_ncs_cnt == 0:
            sub_ncs_cnt = 1
        group_adjacent = lambda a, k: zip(*([iter(a)] * k))
        ncs_group = list(group_adjacent(ncs, sub_ncs_cnt))
        residue_ind = -(len(ncs) % sub_ncs_cnt) if sub_ncs_cnt > 0 else 0
        if residue_ind < 0:
            # 所以如果不能除尽，最终切割的子序列数量为k_split+1, 外部如果需要进行多认为并行，可根据最终切割好的数量重分配任务数
            ncs_group.append(ncs[residue_ind:])
        return ncs_group

    def fit(self, p_ncs=None, threshold=0.65, brust_min=True):
        """
        交易训练集训练拟合函数，根据规则计算gmm分类的范围，使用gmm对训练集交易进行阀值筛选，
        根据brust_min参数决定是否进行全局最优参数计算，最终生成self.rts，self.nts，self.cprs

        eg：self.rts字典形式如下：
            key=component, value=(GaussianMixtured对象，cluster_ind: GMM(component)分component个类中，
            满足子分类的失败率大于阀值threshold的子分类序号)

            {14: (GaussianMixture(max_iter=100, n_components=14, n_init=1),
                Int64Index([7, 8, 9, 10], dtype='int64', name='cluster')),

            15: (GaussianMixture(max_iter=100, n_components=15, n_init=1),
                Int64Index([7, 8, 9, 10], dtype='int64', name='cluster'))}

        eg: self.nts字典对象形式如下所示：
                {
                '14-7':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-11	1	    8.341	        -9.450	        0.730	        12.397	        7	7
                2015-10-28	0	    7.144	        -9.818	        -3.886	        6.955	        39	7
                2015-11-04	0	    12.442	        -10.353	        3.313	        7.840	        41	7
                2016-03-30	0	    13.121	        -8.461	        4.498	        4.070	        49	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        53	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        54	7
                ...............................................................................................

                '14-8':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-26	0	    14.052	        6.061	        7.566	        12.494	        9	8
                2015-04-22	1	    20.640	        2.751	        20.436	        18.781	        23	8
                2015-12-16	0	    12.906	        6.312	        16.638	        12.379	        43	8

                ................................................................................................
                }

        eg: self.cprs pd.DataFrame对象形式如下
                   lcs     lms     lps     lrs
            14_7     6 -0.0552 -0.3310  0.8333
            14_8     3 -0.0622 -0.1866  0.6667
            14_9     1 -0.1128 -0.1128  1.0000
            14_10    2 -0.0327 -0.0654  1.0000
            15_7     4 -0.1116 -0.4465  1.0000
            15_8     3 -0.0622 -0.1866  0.6667
            ...    ...     ...     ...     ...
            27_8     1 -0.1128 -0.1128  1.0000
            27_9     3 -0.0103 -0.0309  0.6667
            27_14    1 -0.0946 -0.0946  1.0000
            27_15    2 -0.1149 -0.2299  1.0000
            27_17    1 -0.1041 -0.1041  1.0000
            27_18    2 -0.0327 -0.0654  1.0000
            27_24    1 -0.0140 -0.0140  1.0000

        :param p_ncs: gmm分类的范围, 可以为具体序列对象如[10, 11, 12....80], 也可以为生成器对象，
                      如xrange(10, 80, 10)，还支持slice对象，eg：slice(10, 80, 10)
        :param threshold: 分类簇中失败率选择的阀值（默认0.65），即大于threshold值的gmm分类簇做为主裁学习目标分类簇
        :param brust_min: bool类型，代表是否进行全局最优参数计算brust_min，如果否则直接使用默认g_brust_min_default(0, 0, 0.65)
                          如果不使用brust_min，即brust_min=False可大大提高训练运行效率
        """
        if self.fiter().df.shape[0] < 50:
            # 有结果的交易最少要有50个，否则直接返回
            self.log_func('order count at least more than 50!')
            return

        if p_ncs is None:
            if self.fiter().df.shape[0] < K_DEFAULT_NCS_MAX:
                # 如果没有设置gmm分类的范围，且总交易单数量小于K_DEFAULT_NCS_MAX，使用默认gmm分类簇范围slice(40, 85, 1)
                ncs = np.arange(K_DEFAULT_NCS_RANG.start,
                                K_DEFAULT_NCS_RANG.stop,
                                K_DEFAULT_NCS_RANG.step)
            else:
                # 交易单数量 > K_DEFAULT_NCS_MAX时，重新计算start，stop，step
                ncs_stop = int(
                    math.ceil(self.fiter().df.shape[0] *
                              K_DEFAULT_NCS_RANG.stop / K_DEFAULT_NCS_MAX))
                # 根据计算出的stop和默认的start重新计算start
                ncs_start = int(ncs_stop - K_DEFAULT_NCS_RANG.start)
                ncs_step = K_DEFAULT_NCS_RANG.step
                ncs = np.arange(ncs_start, ncs_stop, ncs_step)
        elif p_ncs is not None and isinstance(p_ncs, slice):
            # 外部设置的gmm分类的范围为slice切片对象，构造np.arange
            ncs = np.arange(p_ncs.start, p_ncs.stop, p_ncs.step)
        else:
            # 把外部设置的生成器对象如xrange(10, 80, 10)，转换为序列对象，或者tuple, set类型的序列统一到list
            ncs = list(p_ncs)

        if ncs[-1] >= self.fiter().df.shape[0] / 2:
            # 如果分类簇最大分类数ncs[-1]大于总训练集交易量的一半，重新分配gmm分类的范围，使用交易量1/4-1/2
            ncs = np.arange(int(self.fiter().df.shape[0] / 4),
                            int(self.fiter().df.shape[0] / 2))

        # return ncs
        # 这里copy训练集特征对象self.fiter().df，因为要修改df，保持原始训练集特征对象不变
        df = copy.deepcopy(self.fiter().df)
        # 添加一个索引序列，方便之后快速查找原始单据, 即从self.fiter.order_has_ret.iloc[df.ind.values]
        df['ind'] = np.arange(0, df.shape[0])
        """
            eg: df添加ind列
                            result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
            2014-09-24       0          3.378           3.458          3.458
            2014-10-24       0          0.191           2.889          2.809
            2014-10-29       1         -2.026          16.689         -0.761
            2014-10-29       1         -3.427         -11.956         -8.296
            2014-10-29       1         -2.915          39.469         -6.043
                        buy_deg_ang21  ind
            2014-09-24          1.818    0
            2014-10-24         -1.089    1
            2014-10-29          1.980    2
            2014-10-29          6.507    3
            2014-10-29          7.046    4
        """
        clf_component_dict = {}
        cluster_df_dict = {}
        if self.fiter().df.shape[0] < 10000:
            # 交易单总数小于1000个，单进程
            clf_component_dict, cluster_df_dict = _do_gmm_cluster(
                ncs,
                self.fiter().x, df, threshold)
        else:
            n_jobs = g_cpu_cnt
            # 根据进程数切割ncs为n_jobs个子ncs形成ncs_group
            ncs_group = self._sub_ncs_split(ncs, n_jobs)
            parallel = Parallel(n_jobs=len(ncs_group),
                                verbose=0,
                                pre_dispatch='2*n_jobs')

            out = parallel(
                delayed(_do_gmm_cluster)(sub_ncs, self.fiter().x, df,
                                         threshold) for sub_ncs in ncs_group)

            for sub_out in out:
                # 将每一个进程返回的结果进行合并
                clf_component_dict.update(sub_out[0])
                cluster_df_dict.update(sub_out[1])
        self.rts = clf_component_dict
        self.nts = cluster_df_dict
        self.cprs = self._fit_cprs()
        """
         eg: self.cprs形式如
                   lcs     lms     lps     lrs
            14_7     6 -0.0552 -0.3310  0.8333
            14_8     3 -0.0622 -0.1866  0.6667
            14_9     1 -0.1128 -0.1128  1.0000
            14_10    2 -0.0327 -0.0654  1.0000
            15_7     4 -0.1116 -0.4465  1.0000
            15_8     3 -0.0622 -0.1866  0.6667
            ...    ...     ...     ...     ...
            27_8     1 -0.1128 -0.1128  1.0000
            27_9     3 -0.0103 -0.0309  0.6667
            27_14    1 -0.0946 -0.0946  1.0000
            27_15    2 -0.1149 -0.2299  1.0000
            27_17    1 -0.1041 -0.1041  1.0000
            27_18    2 -0.0327 -0.0654  1.0000
            27_24    1 -0.0140 -0.0140  1.0000
        """
        # noinspection PyTypeChecker
        self._fit_brust_min(brust_min)

    def _fit_cprs(self):
        """
        通过self.nts，eg: self.nts字典对象形式如下所示：
                {
                '14-7':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-11	1	    8.341	        -9.450	        0.730	        12.397	        7	7
                2015-10-28	0	    7.144	        -9.818	        -3.886	        6.955	        39	7
                2015-11-04	0	    12.442	        -10.353	        3.313	        7.840	        41	7
                2016-03-30	0	    13.121	        -8.461	        4.498	        4.070	        49	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        53	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        54	7
                ...............................................................................................

                '14-8':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-26	0	    14.052	        6.061	        7.566	        12.494	        9	8
                2015-04-22	1	    20.640	        2.751	        20.436	        18.781	        23	8
                2015-12-16	0	    12.906	        6.312	        16.638	        12.379	        43	8

                ................................................................................................
                }
        计算分类簇中的交易个数，分类簇中的交易的平均收益，分类簇中的交易的总收益比例，
        以及分类簇中的胜率(分类簇中失败亏损的交易数量/类醋中总交易数量), 以上构成字典对象：cprs_dict
            eg: cprs_dict = {'lcs': [], 'lrs': [], 'lps': [], 'lms': []}
                lcs：分类簇中的交易个数
                lrs：代表分类簇中失败亏损的交易数量/类醋中总交易数量
                lps：通过cluster_order_df.profit_cg计算分类簇中的交易的总收益比例
                lms：通过cluster_order_df.profit_cg计算分类簇中的交易的平均收益
        构造pd.DataFrame(cprs_dict, index=cprs_index)提供数据初始

        :param show: 是否可视化lcs，lrs，lps，lms数据(2d, 3d)
        :return: pd.DataFrame对象
                eg: cprs形式如
                       lcs     lms     lps     lrs
                14_7     6 -0.0552 -0.3310  0.8333
                14_8     3 -0.0622 -0.1866  0.6667
                14_9     1 -0.1128 -0.1128  1.0000
                14_10    2 -0.0327 -0.0654  1.0000
                15_7     4 -0.1116 -0.4465  1.0000
                15_8     3 -0.0622 -0.1866  0.6667
                15_9     1 -0.1128 -0.1128  1.0000
                15_10    2 -0.0327 -0.0654  1.0000
                16_7     4 -0.1116 -0.4465  1.0000
                16_8     3 -0.0622 -0.1866  0.6667
                ...    ...     ...     ...     ...
                26_17    1 -0.1041 -0.1041  1.0000
                26_18    2 -0.0327 -0.0654  1.0000
                26_24    1 -0.0140 -0.0140  1.0000
                27_8     1 -0.1128 -0.1128  1.0000
                27_9     3 -0.0103 -0.0309  0.6667
                27_14    1 -0.0946 -0.0946  1.0000
                27_15    2 -0.1149 -0.2299  1.0000
                27_17    1 -0.1041 -0.1041  1.0000
                27_18    2 -0.0327 -0.0654  1.0000
                27_24    1 -0.0140 -0.0140  1.0000

        """
        """
            cprs_dict和cprs_index为了构造pd.DataFrame(cprs_dict, index=cprs_index)提供数据初始
            lcs：分类簇中的交易个数
            lrs：代表分类簇中失败亏损的交易数量/类醋中总交易数量
            lps：通过cluster_order_df.profit_cg计算分类簇中的交易的总收益比例
            lms：通过cluster_order_df.profit_cg计算分类簇中的交易的平均收益
        """

        cprs_dict = {'lcs': [], 'lrs': [], 'lps': [], 'lms': []}
        cprs_index = list()
        for cluster_df_key in self.nts:
            """
                eg: self.nts字典对象形式如下所示
                {
                '14-7':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-11	1	    8.341	        -9.450	        0.730	        12.397	        7	7
                2015-10-28	0	    7.144	        -9.818	        -3.886	        6.955	        39	7
                2015-11-04	0	    12.442	        -10.353	        3.313	        7.840	        41	7
                2016-03-30	0	    13.121	        -8.461	        4.498	        4.070	        49	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        53	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        54	7
                ...............................................................................................

                '14-8':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-26	0	    14.052	        6.061	        7.566	        12.494	        9	8
                2015-04-22	1	    20.640	        2.751	        20.436	        18.781	        23	8
                2015-12-16	0	    12.906	        6.312	        16.638	        12.379	        43	8

                ................................................................................................
                }
            """
            # eg: cluster_df_key = '14-7' 拿出对应的pd.DataFrame对象
            df_cluster = self.nts[cluster_df_key]
            # 分类簇中的交易个数
            cluster_cnt = df_cluster.shape[0]
            """
                eg:
                    df_cluster['result']:

                        2014-11-11    1
                        2015-10-28    0
                        2015-11-04    0
                        2016-03-30    0
                        2016-04-15    0
                        2016-04-15    0

                    df_cluster['result'].value_counts()

                        0    5
                        1    1

                    ->: df_cluster['result'].value_counts()[0] == 5
                    ->: df_cluster['result'].value_counts().sum() == 6

                －> loss_rate = 5 / 6: 代表分类簇中失败亏损的交易数量/类醋中总交易数量
            """
            loss_rate = df_cluster['result'].value_counts(
            )[0] / df_cluster['result'].value_counts().sum()
            # 在fit开始时使用df['ind'] = np.arange(0, df.shape[0])就是为了这里与原始交易单子保留对应关系
            cluster_order_df = self.fiter.order_has_ret.iloc[
                df_cluster.ind.values]
            # 原始交易单子orders_pd中profit_cg数据，代表每笔交易中盈利金额与这笔交易买入金融的比例
            # 通过cluster_order_df.profit_cg计算分类簇中的交易的平均收益
            cluster_profit_mean = cluster_order_df.profit_cg.mean()
            # 通过cluster_order_df.profit_cg计算分类簇中的交易的总收益比例
            cluster_profit_sum = cluster_order_df.profit_cg.sum()

            cprs_dict['lcs'].append(cluster_cnt)
            cprs_dict['lrs'].append(loss_rate)
            cprs_dict['lms'].append(cluster_profit_mean)
            cprs_dict['lps'].append(cluster_profit_sum)
            # cluster_df_key为最终生成pd.DataFrame对象index序列
            cprs_index.append(cluster_df_key)

        cprs = pd.DataFrame(cprs_dict, index=cprs_index)
        """
            eg: cprs形式如
                   lcs     lms     lps     lrs
            14_7     6 -0.0552 -0.3310  0.8333
            14_8     3 -0.0622 -0.1866  0.6667
            14_9     1 -0.1128 -0.1128  1.0000
            14_10    2 -0.0327 -0.0654  1.0000
            15_7     4 -0.1116 -0.4465  1.0000
            15_8     3 -0.0622 -0.1866  0.6667
            15_9     1 -0.1128 -0.1128  1.0000
            15_10    2 -0.0327 -0.0654  1.0000
            16_7     4 -0.1116 -0.4465  1.0000
            16_8     3 -0.0622 -0.1866  0.6667
            ...    ...     ...     ...     ...
            26_17    1 -0.1041 -0.1041  1.0000
            26_18    2 -0.0327 -0.0654  1.0000
            26_24    1 -0.0140 -0.0140  1.0000
            27_8     1 -0.1128 -0.1128  1.0000
            27_9     3 -0.0103 -0.0309  0.6667
            27_14    1 -0.0946 -0.0946  1.0000
            27_15    2 -0.1149 -0.2299  1.0000
            27_17    1 -0.1041 -0.1041  1.0000
            27_18    2 -0.0327 -0.0654  1.0000
            27_24    1 -0.0140 -0.0140  1.0000
        """
        return cprs

    def _fit_brust_min(self, brust_min):
        """
        通过全局最优函数brust_min计算lrs，lps，lms的最优组合，针对最优结果effect_num
        进行如果effect_num == 0使用默认g_brust_min_default(0, 0, 0.65)，通过最优组合
        值对self.cprs对分类簇集合进行筛选形成self.llps
        :param brust_min: bool类型，代表是否进行全局最优参数计算brust_min，如果否则直接使用默认g_brust_min_default(0, 0, 0.65)
                          如果不使用brust_min，即brust_min=False可大大提高训练运行效率
        """
        if self.cprs.shape[0] < 30:
            self.llps = self.cprs[self.cprs.lps < 0]
            self.llps_brust_min = g_brust_min_default
            self.log_func('{}: cprs shape < 30!'.format(
                self.__class__.__name__))
            return
        effect_num = 0
        if brust_min:
            # 使用全局最优参数计算brust_min
            brust_min = self.brust_min()
            _, effect_num = self.min_func(brust_min)

        if effect_num == 0:
            # 如不使用全局最优参数计算brust_min，或者计算出的effect_num＝0使用g_brust_min_default
            brust_min = g_brust_min_default
        # 通过最优参数对分类簇集合进行筛选
        self.llps = self.cprs[(self.cprs['lps'] <= brust_min[0])
                              & (self.cprs['lms'] <= brust_min[1]) &
                              (self.cprs['lrs'] >= brust_min[2])]
        # 将最终从cprs筛选llps使用的brust_min值保存起来
        self.llps_brust_min = brust_min

    def brust_min(self):
        """
        在fit操作之后对self.cprs进行分析可以发现，这些分类中存在很多分类簇中的交易胜率不高，
        但是交易获利比例总和却为正值，也就是说这个分类簇由于非均衡赔率使得非均衡胜率得以保持平衡，
        并且最终获利，那么我们将所有分类簇保存在本地，对之后的交易进行裁决显然是不妥当的，所以需要使用
        全局最优技术对分类簇集合进行筛选

        1. 将lps的范围选定(分类簇中交易获利比例总和最小值-0), 即醋中交易获胜比例组合大于0的过滤
        2. 将lms的范围选定(分类簇中交易获利比例平均值最小值-0), 即醋中交易获利比例平均值大于0的过滤
        3. 将lrs交易醋失败率从最小值－最大值
        4. 计算最小值-0(最大值)之间合适的step值
        5. bnds = (slice(lps_min, 0, lps_step), slice(lms_min, 0, lms_step), slice(lrs_min, lrs_max, lrs_step))
        6. sco.brute中设定最优函数self.min_func_improved使用bnds进行计算
        更多详见brust_min实现函数
        """
        # noinspection PyProtectedMember
        factor_step = 3  #1 #if UltronEnv._g_enable_example_env_ipython else 3
        factor_step *= g_brust_min_step_cnt
        # 分类簇中交易获利比例总和最小值
        lps_min = round(self.cprs['lps'].min(), 3)
        # 默认从最小值－0之间step g_brust_min_step_cnt个：eg: (0 - -13.5) / 100 = 0.135
        lps_step = (0 - lps_min) / factor_step
        # 调整step的粒度保持在合理的范围内，不然太慢，或者粒度太粗
        if lps_step < 0.01 and lps_min + 0.01 <= 0:
            lps_step = 0.01

        # 分类簇中交易获利比例平均值最小值
        lms_min = round(self.cprs['lms'].min(), 3)
        # 默认从最小值－0之间step g_brust_min_step_cnt个：eg: (0 - -0.07) / 100 = 0.0007
        lms_step = (0 - lms_min) / factor_step
        # 调整step的粒度保持在合理的范围内，eg: 0.0007就会被修正到0.01
        if lms_step < 0.01 and lms_step + 0.01 <= 0:
            lms_step = 0.01

        # 交易簇失败率从最小值－最大值
        lrs_min = round(self.cprs['lrs'].min(), 3)
        lrs_max = round(self.cprs['lrs'].max(), 3)
        if lrs_min == lrs_max and lrs_min > 0:
            lrs_min -= 0.05
        ###
        # eg: (0.87 - 0.65) / 100 = 0.002
        lrs_step = (lrs_max - lrs_min) / factor_step
        if lrs_step < 0.05 < lrs_max - lrs_min:
            lrs_step = 0.05

        bnds = (slice(lps_min, 0, lps_step), slice(lms_min, 0, lms_step),
                slice(lrs_min, lrs_max, lrs_step))
        """
            eg: bnds形如
            slice(-13.75, 0, 0.5), slice(-0.07, 0, 0.01), slice(0.65, 0.78, 0.1)

            bnds在最优函数sco.brute内部会使用如np.arange形式展开，即
            eg:
                slice(-13.75, 0, 0.5) －> np.arange(-13.75, 0, 0.5):

                array([-13.75, -13.25, -12.75, -12.25, -11.75, -11.25, -10.75, -10.25,
                        -9.75,  -9.25,  -8.75,  -8.25,  -7.75,  -7.25,  -6.75,  -6.25,
                        -5.75,  -5.25,  -4.75,  -4.25,  -3.75,  -3.25,  -2.75,  -2.25,
                        -1.75,  -1.25,  -0.75,  -0.25])
        """

        progress = 1
        for bnds_pos in (0, 1, 2):
            progress *= len(
                np.arange(bnds[bnds_pos].start, bnds[bnds_pos].stop,
                          bnds[bnds_pos].step))
        # 进行最优时使用的进度条
        self.brust_progress = Progress(
            progress, 0,
            '{}: brute min progress'.format(self.__class__.__name__))
        # 为提高运行效率，不用每次都使用_calc_llps_improved计算
        self.brust_cache = dict()
        brust_result = sco.brute(self.min_func_improved,
                                 bnds,
                                 full_output=False,
                                 finish=None)
        return brust_result

    def min_func(self, l_pmr):
        """
        使用lps，lms，lrs的特点组合值获取crps的子pd.DataFrame对象：
                llps = self.cprs[(self.cprs['lps'] <= l_pmr[0]) & (self.cprs['lms'] <= l_pmr[1]) &
                         (self.cprs['lrs'] >= l_pmr[2])]

            eg: llps形式如

                  lcs     lms     lps  lrs
            15_9    4 -0.1116 -0.4465  0.65
            16_7    12 -0.1786 -0.8122  0.72
            ................................
        通过llps从self.nts字典中获取原始训练集特征数据：
                                result  deg_ang21  deg_ang42  deg_ang60  deg_ang252    ind        cluster
                2011-09-21       0      8.645     -3.294     -8.426      -8.758     40     48      -0.083791
                2011-09-30       0     11.123     -1.887     -2.775       1.585    158     49      -0.066309

        计算llps组合的情况下拦截的单子的数量，以及胜率，失败率，通过：
                improved = (effect_num / self.fiter.order_has_ret.shape[0]) * (loss_rate - win_rate)

                eg：(effect_num / self.fiter.order_has_ret.shape[0]) * (loss_rate - win_rate)
                effect_num ＝ 10, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.8， win_rate ＝0.2
                －> (10 / 100) * (0.8 - 0.2) = 0.06

                effect_num ＝ 50, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.8， win_rate ＝0.2
                －> (50 / 100) * (0.8 - 0.2) = 0.3

                effect_num ＝ 50, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.9， win_rate ＝0.1
                －> (50 / 100) * (0.9 - 0.1) = 0.4

                即最终提升的比例和llps下被拦截的交易数量成正比，和llps下被拦截的交易失败率成正比
        即计算llps组合的情况下，即参数l_pmr给定的lps，lms，lrs的特点组合下的交易提升值
        :param l_pmr: 可迭代序列，eg：(-0.45, -0.11, 0.77), 由三个float值组成的序列
                      l_pmr[0]: 切取self.cprs使用的分类簇交易获利和值，即lps阀值：self.cprs['lps'] <= l_pmr[0]
                      l_pmr[1]: 切取self.cprs使用的分类簇交易获利平均值，即lms阀值：self.cprs['lms'] <= l_pmr[1]
                      l_pmr[2]: 切取self.cprs使用的分类簇交易失败比例，即lrs阀值：self.cprs['lrs'] >= l_pmr[2]
        :return: [improved, effect_num]
        """

        # 组合参数对训练集交易带来的提高默认=-np.inf，因为min_func_improved中的-self.min_func(l_pmr)[0]
        improved = -np.inf
        # 组合参数对训练集影响交易数量默认=0
        effect_num = 0
        # 只有失败率条件是取 lrs >=，获利和lps与平均获利lms取 <=，选取cprs子集llps
        llps = self.cprs[(self.cprs['lps'] <= l_pmr[0])
                         & (self.cprs['lms'] <= l_pmr[1]) &
                         (self.cprs['lrs'] >= l_pmr[2])]
        """
            llps与cprs形式完全一致，llps是cprs的子集pd.DataFrame对象
            eg: llps形式如

                      lcs     lms     lps  lrs
                15_9    4 -0.1116 -0.4465  0.65
                16_7    12 -0.1786 -0.8122  0.72
                ................................
        """
        if llps.empty:
            # 如果筛选出来的子集是空的，说明对整体效果没有提高np.array([0, 0])
            return [improved, effect_num]

        hash_llps = str(llps.index.tolist())
        if hash_llps in self.brust_cache:
            # 为提高运行效率，不用每次都使用_calc_llps_improved计算
            return self.brust_cache[hash_llps]

        # 通过self._calc_llps_improved(llps)继续计算improved和effect_num
        improved, effect_num, _, _ = self._calc_llps_improved(llps)
        brust = [improved, effect_num]
        # 使用缓存字典，llps.index做为key, _calc_llps_improved返回的[improved, effect_num]做为value
        self.brust_cache[hash_llps] = brust
        return brust

    def min_func_improved(self, l_pmr):
        """
        包装min_func具体现实函数，在brust_min函数中使用：
        sco.brute(self.min_func_improved, bnds, full_output=False, finish=None)
        计算最优参数组合，因为在self.min_func中计算的是提高improved值，要想得到最大
        提高参数组合，即self.min_func返回值最大的组合，使用最优sco.brute的目标是
        最小值，所以使用 -self.min_func(l_pmr)[0]，找到最小值，结果的参数组合即为
        最大提高improved值的参数组合
        :param l_pmr: 可迭代序列，eg：(-0.45, -0.11, 0.77), 由三个float值组成的序列
                      l_pmr[0]: 切取self.cprs使用的分类簇交易获利和值，即lps阀值：self.cprs['lps'] <= l_pmr[0]
                      l_pmr[1]: 切取self.cprs使用的分类簇交易获利平均值，即lms阀值：self.cprs['lms'] <= l_pmr[1]
                      l_pmr[2]: 切取self.cprs使用的分类簇交易失败比例，即lrs阀值：self.cprs['lrs'] >= l_pmr[2]
        :return: -self.min_func(l_pmr)[0]，即[improved, effect_num][0], 即improved值，float
        """
        self.brust_progress.show()
        return -self.min_func(l_pmr)[0]

    def _calc_llps_improved(self, llps):
        """
        通过llps从self.nts字典中获取原始训练集特征数据，计算llps组合的情况下拦截的单子的数量，以及胜率，失败率
        即计算llps组合的情况下，即参数l_pmr给定的lps，lms，lrs的特点组合下的交易提升值
        :param llps: eg: llps形式如

                          lcs     lms     lps  lrs
                    15_9    4 -0.1116 -0.4465  0.65
                    16_7    12 -0.1786 -0.8122  0.72
                    ..............................
        :return: improved, effect_num, loss_rate, nts_pd
        """
        nts_pd = pd.DataFrame()
        for component_cluster in llps.index:
            # component_cluster eg:  '14-7', self.nts[component_cluster]即对应的pd.DataFrame对象
            nts_pd = nts_pd.append(self.nts[component_cluster])
            """
                eg: self.nts字典中元素如下所示：
                '14-7':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-11	1	    8.341	        -9.450	        0.730	        12.397	        7	7
                2015-10-28	0	    7.144	        -9.818	        -3.886	        6.955	        39	7
                2015-11-04	0	    12.442	        -10.353	        3.313	        7.840	        41	7
                2016-03-30	0	    13.121	        -8.461	        4.498	        4.070	        49	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        53	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        54	7
            """
        """
            所有的分类簇交易通过nts_pd.append(self.nts[component_cluster])组成必然有重复
            在所有component_cluster中通过nts_pd.drop_duplicates去除重复的单子，注意使用'
            subset='ind'进行去重复，因为self.nts中的pd.DataFrame对象中cluster是不一样的
        """
        nts_pd = nts_pd.drop_duplicates(subset='ind', keep='last')
        """
        nts_pd如下形式，实际上就是所有可能被阻拦的component_cluster对应的交易去除重复的结果
                         result  deg_ang21  deg_ang42  deg_ang60  deg_ang252    ind        cluster
        2011-09-21       0      8.645     -3.294     -8.426      -8.758     40     48      -0.083791
        2011-09-30       0     11.123     -1.887     -2.775       1.585    158     49      -0.066309
        """
        # llps下所有可能被拦截的交易个数
        effect_num = nts_pd.shape[0]
        # llps下被拦截的交易失败率
        loss_rate = nts_pd.result.value_counts(
        )[0] / nts_pd.result.value_counts().sum()
        win_rate = 1 - loss_rate
        # 按照比例有可能提升的效果
        improved = (effect_num /
                    self.fiter.order_has_ret.shape[0]) * (loss_rate - win_rate)
        """
            eg：(effect_num / self.fiter.order_has_ret.shape[0]) * (loss_rate - win_rate)
                effect_num ＝ 10, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.8， win_rate ＝0.2
                －> (10 / 100) * (0.8 - 0.2) = 0.06

                effect_num ＝ 50, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.8， win_rate ＝0.2
                －> (50 / 100) * (0.8 - 0.2) = 0.3

                effect_num ＝ 50, self.fiter.order_has_ret.shape[0] ＝ 100
                loss_rate ＝ 0.9， win_rate ＝0.1
                －> (50 / 100) * (0.9 - 0.1) = 0.4

                即最终提升的比例和llps下被拦截的交易数量成正比，和llps下被拦截的交易失败率成正比
        """
        # TODO 添加其它计算improved的方式，或者权重胜率等多因素决策improved值
        return improved, effect_num, loss_rate, nts_pd

    def dump_file_fn(self):
        """
            主裁本地缓存的存储路径规则：
            UltronEnv.g_project_data_dir ＋ 'ump/ump_main_' ＋ market_name + self.class_unique_id()
        """

        # TODO 如果有裁判覆盖，保留备份，显示通知
        model_path = os.environ[
            'UMP_MODEL_PATH'] if 'UMP_MODEL_PATH' in os.environ else g_project_data_dir
        unique_ump_name = 'ump/ump_main_{}_{}'.format(self.market_name,
                                                      self.class_unique_id())
        return os.path.join(model_path, unique_ump_name)

    def dump_clf(self, llps=None):
        """
        1. 通过llps.index，从self.rts中获取GaussianMixture对象，即self.rts[clf][0]

        eg: llps.index
        Index(['14_7', '14_8', '14_9', '14_10', '15_7', '15_8', '15_9', '15_10',
               '16_7', '16_8', '16_9', '16_10', '17_7', '17_8', '17_9', '17_10',
                ................................................................
               '27_14', '27_15', '27_17', '27_18', '27_24'],
               dtype='object')

        2. 将cluster和GaussianMixture对象做为value，llps.index中的元素做为key形成最终序列化的dict对象形式如下所示
        eg:
            {
            '14_7': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                    means_init=None, n_components=14, n_init=1, precisions_init=None,
                                    random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                    verbose_interval=10, warm_start=False, weights_init=None), 7),
            '14_8': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                    means_init=None, n_components=14, n_init=1, precisions_init=None,
                                    random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                    verbose_interval=10, warm_start=False, weights_init=None), 8)
            ....................................................................................................
            }

        3. 通过UltronFileUtil.dump_pickle最终将元素序列化保存在本地
        :param llps: 需要最终保存的llps对象，默认参数llps=None将使用llps = self.llps
                    eg: llps形式如

                          lcs     lms     lps  lrs
                    15_9    4 -0.1116 -0.4465  0.65
                    16_7    12 -0.1786 -0.8122  0.72
                    ................................
        """
        if llps is None:
            if not hasattr(self, 'llps'):
                # fit中订单不足量等终止情况
                return
            llps = self.llps

        clf_cluster_dict = {}
        for clf_cluster in llps.index:
            """

                eg: llps.index
                Index(['14_7', '14_8', '14_9', '14_10', '15_7', '15_8', '15_9', '15_10',
                       '16_7', '16_8', '16_9', '16_10', '17_7', '17_8', '17_9', '17_10',
                        ................................................................
                       '27_14', '27_15', '27_17', '27_18', '27_24'],
                       dtype='object')
            """
            #  clf, cluster = ('14', '7')
            clf, cluster = clf_cluster.split('_')
            clf = int(clf)
            cluster = int(cluster)
            """
                eg:
                    self.rts[clf][0]
                    GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                    means_init=None, n_components=14, n_init=1, precisions_init=None,
                                    random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                    verbose_interval=10, warm_start=False, weights_init=None)
            """
            clf_cluster_dict[clf_cluster] = (self.rts[clf][0], cluster)
            """
                eg: clf_cluster_dict形式如下所示：
                {
                '14_7': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 7),
                '14_8': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 8)
                ....................................................................................................
                }
            """
        # 通过UltronFileUtil.dump_pickle将clf_cluster_dict进行序列化
        file_util.dump_pickle(clf_cluster_dict,
                              self.dump_file_fn(),
                              how='zero')

    def predict(self, x, need_hit_cnt=1):
        """
        主交易决策函数，从CachedUmpManager中获取缓存clf_cluster_dict，迭代ump字典对象，
        对交易特征形成的x使用字典中元素(clf, cluster)中的clf进行predict
        结果和 (clf, cluster)中存储的cluster一致的情况下代表hit，最终对交易
        做出决策

        :param x: 交易特征形成的x eg: array([[  8.341,  -9.45 ,   0.73 ,  12.397]])
        :param need_hit_cnt: 对交易特征形成的x使用字典中元素(clf, cluster)中的clf进行predict结果和
                            (clf, cluster)中存储的cluster一致的情况下代表hit一次：count_hit += 1
                            只有当need_hit_cnt == count_hit才最终做出决策对交易进行拦截

        :return: 最终做出决策对交易是否进行拦截，拦截即返回1，放行即返回0
        """

        # 统一从CachedUmpManager中获取缓存ump，没有缓存的情况下load_pickle
        clf_cluster_dict = MainBase.dump_clf_manager.get_ump(self)
        """
            eg: clf_cluster_dict
                {
                '14_7': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 7),
                '14_8': (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 8)
                ....................................................................................................
                }
        """
        count_hit = 0
        for clf, cluster in clf_cluster_dict.values():
            """
                eg: (clf, cluster)
                    (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 7)
            """
            predict_cluster = clf.predict(x)
            if predict_cluster == cluster:
                # 即使用clf对x进行predict结果和 (clf, cluster)中存储的cluster一致的情况下代表hit
                count_hit += 1
                if need_hit_cnt == count_hit:
                    # 只有当need_hit_cnt == count_hit才最终做出决策对交易进行拦截，即返回1
                    return 1
        return 0

    def predict_kwargs(self, w_col=None, need_hit_cnt=1, **kwargs):
        """
        主裁交易决策函数，对kwargs关键字参数所描述的交易特征进行拦截决策，从子类对象必须实现的虚方法get_predict_col中获取特征列，
        将kwargs中的特征值转换为x，套接self.predict进行核心裁决判定

        :param w_col: 自定义特征列，一般不会使用，正常情况从子类对象必须实现的虚方法get_predict_col中获取特征列
        :param need_hit_cnt: 透传给self.predict中need_hit_cnt参数，做为分类簇匹配拦截阀值
        :param kwargs: 需要和子类对象实现的虚方法get_predict_col中获取特征列对应的
                       关键字参数，eg: buy_deg_ang42=3.378, buy_deg_ang60=3.458
                                     buy_deg_ang21=3.191, buy_deg_ang252=1.818
        :return: 是否对kwargs关键字参数所描述的交易特征进行拦截，int，不拦截: 0，拦截: 1
        """
        if w_col is None:
            # 如果不自定义特征列，即从子类对象必须实现的虚方法get_predict_col中获取特征列
            w_col = self.get_predict_col()
            """eg, w_col: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']"""

        for col in w_col:
            if col not in kwargs:
                # 如果kwargs中没有某一个特征值，raise ValueError
                raise ValueError('{0} col not in kwargs!'.format(col))
        # 将kwargs中的特征值转换为x
        x = np.array([kwargs[col] for col in w_col])
        """eg, x: array([ 3.378,  3.458,  3.191,  1.818])"""
        x = x.reshape(1, -1)
        # 套接self.predict进行核心裁决判定
        return self.predict(x, need_hit_cnt)

    def choose_cprs_component(self, llps=None):
        """
        辅助工具函数，输出显示在llps下训练集交易中的生效数量，和训练集可提升数据比例
        以及可视化所有训练集中被拦截的交易profit_cg的cumsum

        :param llps: llps对象，默认参数llps=None将使用llps = self.llps

                eg: llps形式如

                      lcs     lms     lps  lrs
                15_9    4 -0.1116 -0.4465  0.65
                16_7    12 -0.1786 -0.8122  0.72
                ................................
        """
        if llps is None:
            llps = self.llps
        improved, effect_num, loss_rate, nts_pd = self._calc_llps_improved(
            llps)
        self.log_func('拦截的交易中正确拦截比例:{0:.3f}'.format(loss_rate))
        self.log_func('训练集中拦截生效后可提升比例:{0:.3f}'.format(improved))
        self.log_func('训练集中拦截生效数量{0}， 训练集中拦截生效占总训练集比例{1:.3f}%'.format(
            effect_num, effect_num / self.fiter.df.shape[0] * 100))
        """
        nts_pd如下形式，实际上就是所有可能被阻拦的component_cluster对应的交易去除重复的结果
                         result  deg_ang21  deg_ang42  deg_ang60  deg_ang252    ind        cluster
        2011-09-21       0      8.645     -3.294     -8.426      -8.758     40     48      -0.083791
        2011-09-30       0     11.123     -1.887     -2.775       1.585    158     49      -0.066309
        """
        # 通过nts_pd通过apply迭代每一行，即每一笔交易的ind映射的原始交易单self.fiter.order_has_ret中的profit_cg值
        nts_pd['profit_cg'] = nts_pd.apply(
            lambda x: self.fiter.order_has_ret.ix[int(x.ind)].profit_cg,
            axis=1)
        """
            eg：nts_pd添加了新列profit_cg后如下所示：
                        result  buy_deg_ang42  buy_deg_ang252  buy_deg_ang60  \
            2015-10-28       0          7.144          -9.818         -3.886
            2015-04-22       1         20.640           2.751         20.436
            2015-06-09       0         10.741          16.352         28.340
            2014-11-11       1          8.341          -9.450          0.730
            2015-11-04       0         12.442         -10.353          3.313
            2016-03-30       0         13.121          -8.461          4.498
            2014-11-26       0         14.052           6.061          7.566
            2016-04-15       0          4.238         -13.247          4.693
            2016-04-15       0          4.238         -13.247          4.693
            2015-12-16       0         12.906           6.312         16.638
            2016-01-29       0         -5.578          16.161         -5.167
            2016-01-29       0         -5.578          16.161         -5.167
            2014-11-12       0          3.963           6.595         -7.524

                        buy_deg_ang21  ind  cluster  profit_cg
            2015-10-28          6.955   39        7    -0.0702
            2015-04-22         18.781   23        8     0.0121
            2015-06-09         -0.937   26        8    -0.1128
            2014-11-11         12.397    7        9     0.1857
            2015-11-04          7.840   41        9    -0.1303
            2016-03-30          4.070   49        9    -0.0863
            2014-11-26         12.494    9       14    -0.0946
            2016-04-15          1.162   53       15    -0.1149
            2016-04-15          1.162   54       15    -0.1149
            2015-12-16         12.379   43       17    -0.1041
            2016-01-29         -3.855   45       18    -0.0327
            2016-01-29         -3.855   46       18    -0.0327
            2014-11-12          6.671    8       24    -0.0140
        """

    def best_hit_cnt_info(self, llps=None):
        """
        辅助工具函数，通过统计在各个分类簇中重复出现的交易单子，计算出
        平均值

        :param llps: llps对象，默认参数llps=None将使用llps = self.llps

                eg: llps形式如
                      lcs     lms     lps  lrs
                15_9    4 -0.1116 -0.4465  0.65
                16_7    12 -0.1786 -0.8122  0.72
                ................................
        """
        if llps is None:
            llps = self.llps

        nts_pd = pd.DataFrame()
        for component_cluster in llps.index:
            # component_cluster eg:  '14-7', self.nts[component_cluster]即对应的pd.DataFrame对象
            nts_pd = nts_pd.append(self.nts[component_cluster])
            """
                eg: self.nts字典中元素如下所示：
                '14-7':
                            result	buy_deg_ang42	buy_deg_ang252	buy_deg_ang60	buy_deg_ang21	ind	cluster
                2014-11-11	1	    8.341	        -9.450	        0.730	        12.397	        7	7
                2015-10-28	0	    7.144	        -9.818	        -3.886	        6.955	        39	7
                2015-11-04	0	    12.442	        -10.353	        3.313	        7.840	        41	7
                2016-03-30	0	    13.121	        -8.461	        4.498	        4.070	        49	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        53	7
                2016-04-15	0	    4.238	        -13.247	        4.693	        1.162	        54	7
            """
        """
            这里不去除重复nts_pd，就是需要统计nts_pd重复ind在各个分类簇中出现的频率
            使用： ind_result_xt = pd.crosstab(nts_pd.ind, nts_pd.result)

            eg：nts_pd.ind
                            2014-11-11     7
                            2015-10-28    39
                            2015-11-04    41
                            2016-03-30    49
                            2016-04-15    53
                            2016-04-15    54
                            2014-11-26     9
                            2015-04-22    23
                            2015-12-16    43
                            2015-06-09    26
                                          ..
                            2014-11-11     7
                            2015-11-04    41
                            2016-03-30    49
                            2014-11-26     9
                            2016-04-15    53
                            2016-04-15    54
                            2015-12-16    43
                            2016-01-29    45
                            2016-01-29    46
                            2014-11-12     8
            eg：nts_pd.result
                            2014-11-11    1
                            2015-10-28    0
                            2015-11-04    0
                            2016-03-30    0
                            2016-04-15    0
                            2016-04-15    0
                            2014-11-26    0
                            2015-04-22    1
                            2015-12-16    0
                            2015-06-09    0
                                         ..
                            2014-11-11    1
                            2015-11-04    0
                            2016-03-30    0
                            2014-11-26    0
                            2016-04-15    0
                            2016-04-15    0
                            2015-12-16    0
                            2016-01-29    0
                            2016-01-29    0
                            2014-11-12    0
            eg：ind_result_xt
                            ind_result_pd
                            result   0  1
                            ind
                            7        0  8
                            8        6  0
                            9       14  0
                            23       0  3
                            26      14  0
                            39       1  0
                            41      14  0
                            43      14  0
                            45      14  0
                            46      14  0
                            49      14  0
                            53      14  0
                            54      14  0
            即ind_result_xt统计出各个分类簇中重复的拦截次数
        """
        ind_result_xt = pd.crosstab(nts_pd.ind, nts_pd.result)

        # 得到失败的平均数, 亦可以使用中位数等其它统计函数
        mean_hit_failed = ind_result_xt[0].mean()
        # 向上取整数，如果是2.1则3
        mean_hit_failed = np.ceil(mean_hit_failed)
        self.log_func('mean_hit_failed = {}'.format(mean_hit_failed))

    def hit_cnt(self, x):
        """
        辅助统计工具函数，从CachedUmpManager中获取缓存clf_cluster_dict，迭代ump字典对象，
        对交易特征形成的x使用字典中元素(clf, cluster)中的clf进行predict
        结果和 (clf, cluster)中存储的cluster一致的情况下代表hit

        :param x: 交易特征形成的x eg: array([[  8.341,  -9.45 ,   0.73 ,  12.397]])
        :return: kwargs关键字参数所描述的交易特征进行分类簇命中统计，返回int值
        """
        # 统一从CachedUmpManager中获取缓存ump，没有缓存的情况下load_pickle
        clf_cluster_dict = Base.dump_clf_manager.get_ump(self)
        hit_cnt = 0
        for clf, cluster in clf_cluster_dict.values():
            """
                eg: (clf, cluster)
                    (GaussianMixture(covariance_type='full', init_params='kmeans', max_iter=100,
                                        means_init=None, n_components=14, n_init=1, precisions_init=None,
                                        random_state=3, reg_covar=1e-06, tol=0.001, verbose=0,
                                        verbose_interval=10, warm_start=False, weights_init=None), 7)
            """
            predict_cluster = clf.predict(x)
            if predict_cluster == cluster:
                # 即使用clf对x进行predict结果和 (clf, cluster)中存储的cluster一致的情况下代表hit
                hit_cnt += 1
        # 最终返回交易特征进行分类簇命中统计结果，int值
        return hit_cnt

    def predict_hit_kwargs(self, w_col=None, **kwargs):
        """
        辅助统计工具函数，对kwargs关键字参数所描述的交易特征进行ump分类簇命中统计，从子类对象必须实现的虚方法get_predict_col中获取特征列，
        将kwargs中的特征值转换为x，套接self.hit_cnt进行分类簇命中统计

        :param w_col: 自定义特征列，一般不会使用，正常情况从子类对象必须实现的虚方法get_predict_col中获取特征列
        :param kwargs: 需要和子类对象实现的虚方法get_predict_col中获取特征列对应的
                       关键字参数，eg: buy_deg_ang42=3.378, buy_deg_ang60=3.458
                                     buy_deg_ang21=3.191, buy_deg_ang252=1.818
        :return: kwargs关键字参数所描述的交易特征进行分类簇命中统计，返回int值
        """
        if w_col is None:
            # 如果不自定义特征列，即从子类对象必须实现的虚方法get_predict_col中获取特征列
            w_col = self.get_predict_col()
            """eg, w_col: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21']"""

        for col in w_col:
            if col not in kwargs:
                # 如果kwargs中没有某一个特征值，raise ValueError
                raise ValueError('col not in kwargs!')
        # 将kwargs中的特征值转换为x
        x = np.array([kwargs[col] for col in w_col])
        """eg, x: array([ 3.378,  3.458,  3.191,  1.818])"""
        x = x.reshape(1, -1)
        # 与predict_kwargs不同的地方主要就是在这里，使用hit_cnt进行分类簇命中统计
        return self.hit_cnt(x)



# -*- coding: utf-8 -*-
import os, logging, datetime, inspect, time, six
from .singleton import Singleton

logger_dict = {
    'info': logging.INFO,
    'warning': logging.WARNING,
    'critical': logging.CRITICAL,
    'debug': logging.DEBUG,
    'error': logging.ERROR
}


class BaseLogger(object):

    def __init__(self,
                 logger_name,
                 log_level='info',
                 log_format=None,
                 log_dir=None):
        self._init_setting(logger_name=logger_name,
                           log_level=log_level,
                           log_format=log_format,
                           log_dir=log_dir)
        #self.logger = None

    def _init_setting(self,
                      logger_name,
                      log_level='info',
                      log_format=None,
                      log_dir=None):
        default_format = '%(asctime)s - [%(filename)s:%(lineno)d] - %(name)s - %(levelname)s - %(message)s'
        default_path = '/var/log/' if 'KD_LOG_PATH' not in os.environ else os.environ[
            'KD_LOG_PATH']
        log_path = default_path if log_dir is None else log_dir
        log_format = default_format if log_format is None else log_format
        formatter = logging.Formatter(log_format)
        self.logger = logging.getLogger(logger_name)
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        self.logger.addHandler(console)
        self.set_level(log_level)
        self._config_file(log_level=log_level,
                          logger_name=logger_name,
                          log_path=log_path,
                          log_format=log_format)

    def _config_file(self, log_level, logger_name, log_format, log_path=None):
        dir_name = os.path.join(log_path, logger_name)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        filename = datetime.datetime.now().date().strftime('%Y-%m-%d') + '.log'
        print(os.path.join(dir_name, filename))
        logging.basicConfig(level=logger_dict[log_level.lower()],
                            format=log_format,
                            datefmt='%m-%d %H:%M',
                            filename=os.path.join(dir_name, filename),
                            filemode='a')

    def reset(self,
              logger_name,
              log_level='info',
              log_format=None,
              log_dir=None):
        self._init_setting(logger_name=logger_name,
                           log_level=log_level,
                           log_format=log_format,
                           log_dir=log_dir)
        return self

    def handle(self):
        return self.logger

    def set_level(self, log_level):
        self.logger.setLevel(logger_dict[log_level.lower()])

    def info(self, msg):
        self.logger.info(msg)

    def warning(self, msg):
        self.logger.warning(msg)

    def critical(self, msg):
        self.logger.critical(msg)

    def debug(self, msg):
        self.logger.debug(msg)

    def error(self, msg):
        self.logger.error(msg)


class RDLogger(object):

    def __init__(self, logger_name, log_level='info'):
        self._logger_name = logger_name
        self.logger = logging.getLogger(logger_name)
        console = logging.StreamHandler()
        self.logger.addHandler(console)
        self.set_level(log_level)

    def set_level(self, log_level):
        self.logger.setLevel(logger_dict[log_level.lower()])

    def message(self, level, message):
        _, file_name, line_no, function_name, _, _ = inspect.stack()[2]
        return "[%s line:%s] %s - %s - %s - %s" % (
            file_name, line_no,
            time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
            self._logger_name, level.upper(), message)

    def info(self, msg):
        self.logger.info(self.message('info', msg))

    def warning(self, msg):
        self.logger.info(self.message('warning', msg))

    def critical(self, msg):
        self.logger.info(self.message('critical', msg))

    def debug(self, msg):
        self.logger.info(self.message('debug', msg))

    def error(self, msg):
        self.logger.info(self.message('error', msg))


class LoggerFactory(object):

    @classmethod
    def create_engine(self, name='custom'):
        if name == 'rd':
            return RDLogger
        else:
            return BaseLogger(name)


@six.add_metaclass(Singleton)
class CustomLogger():

    def __init__(self, name=None):
        self.logger = LoggerFactory.create_engine(name=name)

    def reset(self, name):
        self.logger = self.logger.reset(logger_name=name)
        global mlc_logger
        mlc_logger = self.handle()

    def handle(self):
        return self.logger.handle()


kd_logger = CustomLogger('ultron' if 'KD_LOG_NAME' not in os.environ else os.
                         environ['ULTRON_LOG_NAME']).handle()



# -*- encoding:utf-8 -*-
"""
中间层，从上层拿到x，y，df
拥有create estimator

"""
import functools
import pandas as pd
import numpy as np
from enum import Enum
from sklearn.preprocessing import label_binarize, StandardScaler, binarize
from ultron.ump.core.fixes import train_test_split, cross_val_score, mean_squared_error_scorer, six
from ultron.ump.core.fixes import six
from ultron.ump.core.fixes import mean_squared_error_scorer
from ultron.ump.model.creater import MLCreater


class _EMLScoreType(Enum):
    """针对有监督学习的度量支持enum"""
    """有监督学习度量准确率"""
    E_SCORE_ACCURACY = 'accuracy'
    """有监督学习度量mse"""
    E_SCORE_MSE = mean_squared_error_scorer
    """有监督学习度量roc_auc"""
    E_SCORE_ROC_AUC = 'roc_auc'


class EMLFitType(Enum):
    """支持常使用的学习器类别enum"""
    """有监督学习：自动选择，根据y的label数量，> 10使用回归否则使用分类"""
    E_FIT_AUTO = 'auto'
    """有监督学习：回归"""
    E_FIT_REG = 'reg'
    """有监督学习：分类"""
    E_FIT_CLF = 'clf'
    """无监督学习：HMM"""
    E_FIT_HMM = 'hmm'
    """无监督学习：PCA"""
    E_FIT_PCA = 'pca'
    """无监督学习：KMEAN"""
    E_FIT_KMEAN = 'kmean'


def entry_wrapper(support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG,
                           EMLFitType.E_FIT_HMM, EMLFitType.E_FIT_PCA,
                           EMLFitType.E_FIT_KMEAN)):
    """
    类装饰器函数，对关键字参数中的fiter_type进行标准化，eg，fiter_type参数是'clf'， 转换为EMLFitType(fiter_type)
    赋予self.fiter_type，检测当前使用的具体学习器不在support参数中不执行被装饰的func函数了，打个log返回

    :param support: 默认 support=(EMLFitType.E_FIT_CLF, EMLFitType.E_FIT_REG, EMLFitType.E_FIT_HMM,
                           EMLFitType.E_FIT_PCA, EMLFitType.E_FIT_KMEAN)
                    即支持所有，被装饰的函数根据自身特性选择装饰参数
    """

    def decorate(func):

        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            org_fiter_type = self.fiter_type
            if 'fiter_type' in kwargs:
                # 如果传递了fiter_type参数，pop出来
                fiter_type = kwargs.pop('fiter_type')
                # 如果传递的fiter_type参数是str，eg：'clf'， 转换为EMLFitType(fiter_type)
                if isinstance(fiter_type, six.string_types):
                    fiter_type = EMLFitType(fiter_type)
                self.fiter_type = fiter_type

            check_support = self.fiter_type
            if self.fiter_type == EMLFitType.E_FIT_AUTO:
                # 把auto的归到具体的分类或者回归
                check_y = self.y
                if 'y' in kwargs:
                    check_y = kwargs['y']
                check_support = EMLFitType.E_FIT_CLF if len(
                    np.unique(check_y)) <= 10 else EMLFitType.E_FIT_REG
            if check_support not in support:
                # 当前使用的具体学习器不在support参数中不执行被装饰的func函数了，打个log返回
                self.log_func('{} not support {}!'.format(
                    func.__name__, check_support.value))
                # 如果没能成功执行把类型再切换回来
                self.fiter_type = org_fiter_type
                return

            return func(self, *args, **kwargs)

        return wrapper

    return decorate


class Estimator(object):
    """封装有简单学习及无监督学习方法以及相关操作类"""

    def __init__(self, x, y, df, fiter_type=EMLFitType.E_FIT_AUTO):
        """
        UltronML属于中间层需要所有原料都配齐，x, y, df，构造方式参考
        create_test_fiter方法中的实行流程

        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param df: 拆分x，y使用的pd.DataFrame对象
        :param fiter_type: 使用的学习器类型，默认使用E_FIT_AUTO即根据y的label数量选择分类或者回归
        """
        self.estimator = MLCreater()
        # 如果传递进来的是字符串类型，转换为EMLFitType
        if isinstance(fiter_type, six.string_types):
            fiter_type = EMLFitType(fiter_type)
        self.x = x
        self.y = y
        self.df = df
        # ipython notebook下使用logging.info
        self.log_func = print  #logging.info if UltronEnv.g_is_ipython else print
        self.fiter_type = fiter_type

    def is_supervised_learning(self):
        """
        返回self.fiter_type所使用的是有监督学习还是无监督学习
        :return: bool，True: 有监督，False: 无监督
        """
        return self.fiter_type == EMLFitType.E_FIT_REG or self.fiter_type == EMLFitType.E_FIT_CLF or \
            self.fiter_type == EMLFitType.E_FIT_AUTO

    def echo_info(self, fiter=None):
        """
        显示fiter class信息，self.df信息包括，head，tail，describe
        eg：
            fiter class is: DecisionTreeClassifier(class_weight=None, criterion='gini', max_depth=None,
                max_features=None, max_leaf_nodes=None,
                min_impurity_split=1e-07, min_samples_leaf=1,
                min_samples_split=2, min_weight_fraction_leaf=0.0,
                presort=False, random_state=None, splitter='best')
            describe:
                          y        x0        x1        x2        x3
            count  150.0000  150.0000  150.0000  150.0000  150.0000
            mean     1.0000    5.8433    3.0540    3.7587    1.1987
            std      0.8192    0.8281    0.4336    1.7644    0.7632
            min      0.0000    4.3000    2.0000    1.0000    0.1000
            25%      0.0000    5.1000    2.8000    1.6000    0.3000
            50%      1.0000    5.8000    3.0000    4.3500    1.3000
            75%      2.0000    6.4000    3.3000    5.1000    1.8000
            max      2.0000    7.9000    4.4000    6.9000    2.5000
        :param fiter:
        :return:
        """
        if fiter is None:
            fiter = self.get_fiter()
        self.log_func('fiter class is: {}'.format(fiter))
        self.log_func('describe:\n{}'.format(self.df.describe()))
        self.log_func('head:\n{}'.format(self.df.head()))
        self.log_func('tail:\n{}'.format(self.df.tail()))

    def get_fiter(self):
        """
        根据self.fiter_type的类型选择从self.estimator返回学习器对象

        self.fiter_type == EMLFitType.E_FIT_AUTO：
            自动选择有简单学习，当y的label数量 < 10个使用分类self.estimator.clf，否则回归self.estimator.reg
        self.fiter_type == EMLFitType.E_FIT_REG:
            使用有监督学习回归self.estimator.reg
        self.fiter_type == EMLFitType.E_FIT_CLF:
            使用有监督学习分类self.estimator.clf
        self.fiter_type == EMLFitType.E_FIT_HMM:
            使用无监督学习hmm，self.estimator.hmm
        self.fiter_type == EMLFitType.E_FIT_PCA:
            使用无监督学习pca，self.estimator.pca
        self.fiter_type == EMLFitType.E_FIT_KMEAN:
            使用无监督学习kmean，self.estimator.kmean
        :return: 返回学习器对象
        """
        if self.fiter_type == EMLFitType.E_FIT_AUTO:
            if len(np.unique(self.y)) <= 10:
                # 小于等于10个class的y就认为是要用分类了
                fiter = self.estimator.clf
            else:
                fiter = self.estimator.reg
        elif self.fiter_type == EMLFitType.E_FIT_REG:
            fiter = self.estimator.reg
        elif self.fiter_type == EMLFitType.E_FIT_CLF:
            fiter = self.estimator.clf
        elif self.fiter_type == EMLFitType.E_FIT_HMM:
            if self.estimator.hmm is None:
                self.estimator.hmm_gaussian()
            fiter = self.estimator.hmm
        elif self.fiter_type == EMLFitType.E_FIT_PCA:
            if self.estimator.pca is None:
                self.estimator.pca_decomposition()
            fiter = self.estimator.pca
        elif self.fiter_type == EMLFitType.E_FIT_KMEAN:
            if self.estimator.kmean is None:
                self.estimator.kmean_cluster()
            fiter = self.estimator.kmean
        else:
            raise TypeError('self.fiter_type = {}, is error type'.format(
                self.fiter_type))

        return fiter

    @entry_wrapper(support=(EMLFitType.E_FIT_CLF, ))
    def cross_val_accuracy_score(self, cv=10, **kwargs):
        """
        被装饰器entry_wrapper(support=(EMLFitType.E_FIT_CLF,))装饰，
        即只支持有监督学习分类，使用cross_val_score对数据进行accuracy度量
        :param cv: 透传cross_val_score的参数，默认10
        :param kwargs: 外部可以传递x, y, 通过
                                x = kwargs.pop('x', self.x)
                                y = kwargs.pop('y', self.y)
                       确定传递self._do_cross_val_score中参数x，y，
                       装饰器使用的fiter_type，eg：ttn_ultron.cross_val_accuracy_score(fiter_type=ml.EMLFitType.E_FIT_CLF)
        :return: cross_val_score返回的score序列，
                 eg: array([ 1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  0.95,  1.  ])
        """
        x = kwargs.pop('x', self.x)
        y = kwargs.pop('y', self.y)
        return self._do_cross_val_score(x, y, cv,
                                        _EMLScoreType.E_SCORE_ACCURACY.value)

    def _do_cross_val_score(self, x, y, cv, scoring):
        """
        封装sklearn中cross_val_score方法， 参数x, y, cv, scoring透传cross_val_score
        :param x: 训练集x矩阵，numpy矩阵
        :param y: 训练集y序列，numpy序列
        :param cv: 透传cross_val_score，cv参数，int
        :param scoring: 透传cross_val_score, 使用的度量方法
        :return: cross_val_score返回的score序列，
                 eg: array([ 1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  1.  ,  0.9 ,  0.95,  1.  ])
        """
        fiter = self.get_fiter()
        """
            eg: fiter
            DecisionTreeClassifier(class_weight=None, criterion='gini', max_depth=None,
                                    max_features=None, max_leaf_nodes=None,
                                    min_impurity_split=1e-07, min_samples_leaf=1,
                                    min_samples_split=2, min_weight_fraction_leaf=0.0,
                                    presort=False, random_state=None, splitter='best')
        """
        if scoring == _EMLScoreType.E_SCORE_ROC_AUC.value and len(
                np.unique(y)) != 2:
            # roc auc的度量下且y的label数量不是2项分类，首先使用label_binarize进行处理
            y_label_binarize = label_binarize(y, classes=np.unique(y))
            """
                eg：
                    np.unique(y) ＝ array([0, 1, 2])
                    y_label_binarize:
                    array([[1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           [1, 0, 0],
                           .........
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           [0, 1, 0],
                           .........
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1],
                           [0, 0, 1]])
            """
            label_cnt = len(np.unique(y))
            # one vs rest的score平均值的和
            mean_sum = 0
            # one vs rest中的最好score平均值
            best_mean = 0
            # 最好score平均值(best_mean)的score序列，做为结果返回
            scores = list()
            for ind in np.arange(0, label_cnt):
                # 开始 one vs rest
                _y = y_label_binarize[:, ind]
                """
                    eg: _y
                    array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                           1, 1, 1, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
                           0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
                """
                tmp_scores = cross_val_score(fiter,
                                             x,
                                             _y,
                                             cv=cv,
                                             scoring=scoring)
                tmp_mean = np.mean(tmp_scores)
                # one vs rest的score平均值进行叠加sum
                mean_sum += tmp_mean
                if len(scores) == 0 or tmp_mean > best_mean:
                    scores = tmp_scores
            # one vs rest的score平均值的和 / label_cnt
            mean_sc = mean_sum / label_cnt
        else:
            scores = cross_val_score(fiter, x, y, cv=cv, scoring=scoring)
            # 计算度量的score平均值，做为log输出，结果返回的仍然是scores
            mean_sc = -np.mean(np.sqrt(-scores)) if scoring == mean_squared_error_scorer \
                else np.mean(scores)
        self.log_func('{} score mean: {}'.format(fiter.__class__.__name__,
                                                 mean_sc))

        return scores



# -*- encoding:utf-8 -*-
"""示例ump边裁特征多混模块"""

from ultron.ump.trade.ml_feature import FeatureDeg, FeaturePrice, FeatureWave, FeatureAtr
from ultron.ump.ump.base import ump_edge_make_xy, BuyUmpMixin
from ultron.ump.ump.edge_base import EdgeBase
from ultron.ump.model.principles import Principles


class FullFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

        在边裁__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        """
         make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
         order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
         make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

        :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
        :return: self.order_has_ret中使用filter选取特征列形成full_df
                 ump_edge_make_xy装饰器在make_xy返回full_df后做转换matrix，形成x，y等工作
        """

        filter_list = ['profit', 'profit_cg']
        cols = FeatureDeg().get_feature_ump_keys(ump_cls=EdgeFull)
        cols.extend(FeaturePrice().get_feature_ump_keys(ump_cls=EdgeFull))
        cols.extend(FeatureWave().get_feature_ump_keys(ump_cls=EdgeFull))
        cols.extend(FeatureAtr().get_feature_ump_keys(ump_cls=EdgeFull))

        filter_list.extend(cols)
        """
            filter_list:
            ['profit', 'profit_cg', 'buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21',
            'buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252', 'buy_wave_score1',
            'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std']
        """
        # noinspection PyUnresolvedReferences
        full_df = self.order_has_ret.filter(filter_list)
        """
            eg: full_df
                           profit  profit_cg  buy_deg_ang42  buy_deg_ang252  \
            2014-09-24  -22618.04    -0.0566          3.378           3.458
            2014-10-24  -29690.28    -0.0742          0.191           2.889
            2014-10-29   18959.19     0.0542         -2.026          16.689
            2014-10-29  148209.36     0.5022         -3.427         -11.956
            2014-10-29   24867.60     0.0952         -2.915          39.469
            2014-10-29   18959.19     0.0542         -2.026          16.689
            ...               ...        ...            ...             ...
            2016-04-13  -28044.40    -0.1159          6.603         -31.459
            2016-04-14  -18645.93    -0.0467          4.611          18.428
            2016-04-15  -32484.79    -0.1149          4.238         -13.247
            2016-04-15  -32484.79    -0.1149          4.238         -13.247
            2016-04-29     290.96     0.0007          1.445          16.266
            2016-04-29     290.96     0.0007          1.445          16.266

                        buy_deg_ang60  buy_deg_ang21  buy_price_rank120  buy_price_rank90  \
            2014-09-24          3.458          1.818              1.000             1.000
            2014-10-24          2.809         -1.089              1.000             1.000
            2014-10-29         -0.761          1.980              1.000             1.000
            2014-10-29         -8.296          6.507              0.925             0.900
            2014-10-29         -6.043          7.046              0.962             0.950
            2014-10-29         -0.761          1.980              1.000             1.000
            ...                   ...            ...                ...               ...
            2016-04-13          0.191          4.457              0.567             0.722
            2016-04-14          3.134          0.733              0.875             0.878
            2016-04-15          4.693          1.162              0.775             0.733
            2016-04-15          4.693          1.162              0.775             0.733
            2016-04-29          4.615         -1.115              1.000             1.000
            2016-04-29          4.615         -1.115              1.000             1.000

                        buy_price_rank60  buy_price_rank252  buy_wave_score1  \
            2014-09-24             1.000              1.000            0.287
            2014-10-24             1.000              1.000            0.596
            2014-10-29             1.000              1.000            0.444
            2014-10-29             0.883              0.750           -0.173
            2014-10-29             0.925              0.982            0.031
            2014-10-29             1.000              1.000            0.444
            ...                      ...                ...              ...
            2016-04-13             1.000              0.270            0.271
            2016-04-14             0.967              0.940           -0.030
            2016-04-15             1.000              0.631            0.596
            2016-04-15             1.000              0.631            0.596
            2016-04-29             1.000              1.000            0.743
            2016-04-29             1.000              1.000            0.743

                        buy_wave_score2  buy_wave_score3  buy_atr_std
            2014-09-24            0.234            0.218        0.226
            2014-10-24            0.488            0.449        0.146
            2014-10-29            0.338            0.329        0.538
            2014-10-29           -0.202           -0.203        0.558
            2014-10-29           -0.128           -0.173        0.063
            2014-10-29            0.338            0.329        0.538
            ...                     ...              ...          ...
            2016-04-13            0.509            0.651        0.262
            2016-04-14            0.081            0.170        0.049
            2016-04-15            0.753            0.800        0.135
            2016-04-15            0.753            0.800        0.135
            2016-04-29            0.840            0.918        0.838
            2016-04-29            0.840            0.918        0.838
        """
        return full_df


class EdgeFull(EdgeBase, BuyUmpMixin):
    """边裁多混特征类，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        边裁多混特征keys：
            ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
            'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
            'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
            'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']

        :return: ['buy_deg_ang42', 'buy_deg_ang252', 'buy_deg_ang60', 'buy_deg_ang21', 'buy_price_rank120',
                'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252',
                'buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3', 'buy_atr_std',
                'buy_diff_down_days', 'buy_jump_up_power', 'buy_diff_up_days', 'buy_jump_down_power']
        """

        cols = FeatureDeg().get_feature_ump_keys(ump_cls=EdgeFull)
        cols.extend(FeaturePrice().get_feature_ump_keys(ump_cls=EdgeFull))
        cols.extend(FeatureWave().get_feature_ump_keys(ump_cls=EdgeFull))
        cols.extend(FeatureAtr().get_feature_ump_keys(ump_cls=EdgeFull))
        return cols

    def get_fiter_class(self):
        """
        边裁多混特征返回的UltronMLPd子类：UltronUmpEdgeFull.UmpFullFiter
        :return: UltronUmpEdgeFull.UmpFullFiter
        """
        return FullFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'full_edge'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'full_edge'


# -*- coding: utf-8 -*-
import datetime
import time
import json
from ultron.cluster.central.extern_modules.base_module import BaseModule
from ultron.cluster.central.extern_modules.middleware.module import ModelsSingleton


class Module(BaseModule):

    def __init__(self, name, redis_client):
        super(Module, self).__init__(name, redis_client)
        self._func = {'upload_packet': self.upload_packet}

    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)

    def upload_packet(self, result):
        second_time = time.time()
        task_id = int(second_time * 1000000 +
                      datetime.datetime.now().microsecond)
        queue_list = ModelsSingleton().get_all_queue()
        for queue in queue_list:
            self._redis_client.hset(queue, str(task_id), json.dumps(result))



# -*- encoding:utf-8 -*-
"""示例ump价格波动特征模块"""

from ultron.ump.trade.ml_feature import FeatureWave
from ultron.ump.ump.base import ump_main_make_xy, BuyUmpMixin
from ultron.ump.ump.main_base import MainBase
from ultron.ump.model.principles import Principles


class WaveFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_main_make_xy装饰

        在__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=self.orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_main_make_xy
    def make_xy(self, **kwarg):
        """
        make_xy通过装饰器ump_main_make_xy进行二次包装
        这里只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df
        :param kwarg: ump_main_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
                      kwargs['scaler']    做为可选关键字参数：控制在make_xy中返回的特征矩阵数据是否进行标准化处理

        :return: self.order_has_ret中使用filter选取特征列形成wave_df
                 ump_main_make_xy装饰器在make_xy返回wave_df后做转换matrix，形成x，y等工作
        """

        # regex=result|buy_wave_score1|buy_wave_score2|buy_wave_score3'
        regex = 'result|{}'.format('|'.join(
            FeatureWave().get_feature_ump_keys(ump_cls=MainWave)))
        # noinspection PyUnresolvedReferences
        wave_df = self.order_has_ret.filter(regex=regex)
        """
             eg: wave_df
                        result  buy_wave_score1  buy_wave_score2  buy_wave_score3
            2014-09-24       0            0.287            0.234            0.218
            2014-10-24       0            0.596            0.488            0.449
            2014-10-29       1            0.444            0.338            0.329
            2014-10-29       1           -0.173           -0.202           -0.203
            2014-10-29       1            0.031           -0.128           -0.173
            2014-10-29       1            0.444            0.338            0.329
            2014-11-03       1            0.018           -0.128           -0.172
            2014-11-11       1           -0.144           -0.060            0.001
            2014-11-12       0           -0.453           -0.505           -0.509
            2014-11-26       0           -0.005           -0.007            0.015
            ...            ...              ...              ...              ...
            2016-03-14       1            0.928            0.941            0.948
            2016-03-14       0            1.209            0.891            0.788
            2016-03-30       0            0.470            0.630            0.702
            2016-04-04       1            0.363            0.608            0.743
            2016-04-13       0            0.271            0.509            0.651
            2016-04-14       0           -0.030            0.081            0.170
            2016-04-15       0            0.596            0.753            0.800
            2016-04-15       0            0.596            0.753            0.800
            2016-04-29       1            0.743            0.840            0.918
            2016-04-29       1            0.743            0.840            0.918
        """

        return wave_df


class MainWave(MainBase, BuyUmpMixin):
    """主裁价格波动特征类，UltronUmpMainBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        主裁价格波动特征keys：['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        :return: ['buy_wave_score1', 'buy_wave_score2', 'buy_wave_score3']
        """

        return FeatureWave().get_feature_ump_keys(ump_cls=MainWave)

    def get_fiter_class(self):
        """
        主裁价格波动特征返回的UltronMLPd子类：UltronUmpMainWave.UmpWaveFiter
        :return: UltronUmpMainWave.UmpWaveFiter
        """
        return WaveFiter

    @classmethod
    def class_unique_id(cls):
        """
        具体ump类关键字唯一名称，类方法：return 'wave_main'
        主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
        具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'wave_main'


# -*- encoding:utf-8 -*-
"""
    技术线对象，对外执行，输出模块
"""
import math
import numpy as np
import pandas as pd
from ultron.kdutils.lazy import LazyFunc
from ultron.ump.core.base import FreezeAttrMixin
from ultron.kdutils.decorator import arr_to_numpy
from ultron.kdutils import regression
"""step_x_to_step函数中序列步长的常数单元值"""
g_step_unit = 10


class Line(FreezeAttrMixin):
    """技术线封装执行对外操作的对象类"""

    def __init__(self, line, line_name, **kwargs):
        """
        :param line: 技术线可迭代序列，内部会通过arr_to_numpy统一转换numpy
        :param line_name: 技术线名称，str对象
        :param kwargs mean: 外部可选择通过kwargs设置mean，如不设置line.mean()
        :param kwargs std: 外部可选择通过kwargs设置std，如不设置line.std()
        :param kwargs high: 外部可选择通过kwargs设置high，如不设置self.mean + self.std
        :param kwargs low: 外部可选择通过kwargs设置low，如不设置self.mean - self.std
        :param kwargs close: 外部可选择通过kwargs设置close，如不设置line[-1]
        """
        # 把序列的nan进行填充，实际上应该是外面根据数据逻辑把nan进行填充好了再传递进来，这里只能都使用bfill填了
        line = pd.Series(line).fillna(method='bfill')
        self.tl = arr_to_numpy(line)
        self.mean = kwargs.pop('mean', self.tl.mean())
        self.std = kwargs.pop('std', self.tl.std())
        self.high = kwargs.pop('high', self.mean + self.std)
        self.low = kwargs.pop('low', self.mean - self.std)
        self.close = kwargs.pop('close', self.tl[-1])

        self.x = np.arange(0, self.tl.shape[0])
        self.line_name = line_name

        for k, v in kwargs:
            # 需要设置什么都通过kwargs设置进来，不然_freeze后无法设置
            setattr(self, k, v)
        # 需要进行定稿，初始化好就不能动
        self._freeze()

    @LazyFunc
    def score(self):
        """
        被LazyFunc装饰：
        score代表当前技术线值在当前的位置， (self.close - self.low) / (self.high - self.low)
        eg：
            self.high ＝ 100， self.low＝0，self.close＝80
            －> (self.close - self.low) / (self.high - self.low) = 0.8
            即代表当前位置在整体的0.8位置上

        :return: 技术线当前score, 返回值在0-1之间
        """
        if self.high == self.low:
            score = 0.8 if self.close > self.low else 0.2
        else:
            score = (self.close - self.low) / (self.high - self.low)
        return score

    @LazyFunc
    def y_zoom(self):
        """
        被LazyFunc装饰：
        获取对象技术线tl被self.x缩放后的序列y_zoom
        :return: 放后的序列y_zoom
        """
        zoom_factor = self.x.max() / self.tl.max()
        y_zoom = zoom_factor * self.tl
        return y_zoom

    def step_x_to_step(self, step_x):
        """
        针对技术线的时间范围步长选择函数，在show_shift_distance，show_regress_trend_channel，
        show_skeleton_channel等涉及时间步长的函数中用来控制步长范围
        :param step_x: 时间步长控制参数，float
        :return: 最终输出被控制在2-len(self.tl), int
        """
        if step_x <= 0:
            # 不正常step_x规范到正常范围中
            #log_func(
            #    'input step_x={} is error, change to step_x=1'.format(step_x))
            step_x = 1
        # 如果需要调整更细的粒度，调整g_step_unit的值
        step = int(math.floor(len(self.tl) / g_step_unit / step_x))
        # 输出被控制在2-len(self.tl)
        step = len(self.tl) if step > len(self.tl) else step
        step = 2 if step < 2 else step
        return step

    def is_up_trend(self, up_deg_threshold=5):
        """
        判断走势是否符合上升走势：
        1. 判断走势是否可以使用一次拟合进行描述
        2. 如果可以使用1次拟合进行描述，计算一次拟合趋势角度
        3. 如果1次拟合趋势角度 >= up_deg_threshold判定上升
        :param up_deg_threshold: 判定一次拟合趋势角度为上升趋势的阀值角度，默认5
        :return: 是否上升趋势
        """
        valid = regression.valid_poly(self.tl, poly=1)
        if valid:
            deg = regression.calc_regress_deg(self.tl)
            if deg >= up_deg_threshold:
                return True
        return False

    def is_down_trend(self, down_deg_threshold=-5):
        """
        判断走势是否符合下降走势：
        1. 判断走势是否可以使用一次拟合进行描述
        2. 如果可以使用1次拟合进行描述，计算一次拟合趋势角度
        3. 如果1次拟合趋势角度 <= down_deg_threshold判定下降
        :param down_deg_threshold: 判定一次拟合趋势角度为下降趋势的阀值角度，默认－5
        :return: 是否下降趋势
        """
        valid = regression.valid_poly(self.tl, poly=1)
        if valid:
            deg = regression.calc_regress_deg(self.tl)
            if deg <= down_deg_threshold:
                return True
        return False


# -*- coding: utf-8 -*-

# 列名: 股票代码 code, 日期 trade_date
import pandas as pd
import numpy as np
import scipy.stats as st
import statsmodels.api as sm 
from sklearn.covariance import LedoitWolf
from cvxopt import matrix, solvers
from . kutil import calc_ic

def factor_combine(factor_df, factor_list, weight_df):
    """
    因子合成
    参数:
        factor_df: DataFrame, 待合成因子值
        factor_list: list, 待合成因子列表
        weight_df: DataFrame, 因子权重
    返回:
        DataFrame, 复合因子
    """
    factor_df = factor_df.dropna(subset=factor_list, how='all')
    
    merge_df = factor_df.merge(weight_df, on=['trade_date'], suffixes=('', '_w'))
    
    w_list = [fn+'_w' for fn in factor_list]
    merge_df[w_list] = np.where(merge_df[factor_list].isnull(), 0, merge_df[w_list])
    merge_df['combine'] = np.sum(merge_df[factor_list].fillna(0).values*merge_df[w_list].values, axis=1)
    merge_df['combine'] = merge_df['combine']/merge_df[w_list].abs().sum(axis=1)
    
    return merge_df[['code', 'trade_date', 'combine']]

def equal_combine(factor_df, factor_list, ):
    """
    等权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        factor_list: list, 待合成因子列表
    返回:
        DataFrame, 复合因子
    """
    factor_df = factor_df.copy()
    return factor_df[factor_list].mean(axis=1).values
    #factor_df['equal']= factor_df[factor_list].mean(axis=1)
    #return factor_df[['code', 'trade_date', 'equal']]


def hist_ret_combine(factor_df, mret_df, factor_list, span,
                     size_indu_df=None, 
                     indu_list=[],
                     method='equal', half_life=1):
    """
    历史收益率加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算历史收益率均值
        method: 历史收益率均值计算方法。'equal':算术平均；'half_life':半衰加权
        half_life: int, 半衰加权时的半衰期
    返回:
        DataFrame, 复合因子
    """
    #merged_df = factor_df.merge(size_indu_df, on=['code', 'trade_date']).merge(mret_df, on=['code', 'trade_date'])
    merged_df = factor_df.merge(mret_df, on=['code', 'trade_date'])
    merged_df = merged_df.dropna(subset=['nxt1_ret'])
    # 计算历史因子收益率
    hist_ret = []
    for fn in factor_list:
        tmp = merged_df.groupby('trade_date').apply(lambda df: sm.OLS(df['nxt1_ret'], sm.add_constant(df[[fn]+indu_list]), missing='drop').fit().params[1])
        hist_ret.append(tmp)
    hist_ret = pd.concat(hist_ret, axis=1)
    hist_ret.columns = factor_list
    hist_ret = hist_ret.reset_index().sort_values('trade_date')
    hist_ret['trade_date'] = hist_ret['trade_date'].shift(-1)
    
    # 计算历史因子收益率移动平均
    for fn in factor_list:
        if method == 'equal':
            hist_ret[fn+'_ma'] = hist_ret[fn].rolling(span).mean()
        elif method == 'half_life':
            hist_ret[fn+'_ma'] = hist_ret[fn].rolling(span).apply(lambda x: np.average(x, weights=list(reversed([0.5**(1.0*i/half_life) for i in range(span)]))))
        else:
            return 0
    
    hist_ret = hist_ret.dropna()
    hist_ret = hist_ret.drop(factor_list, axis=1)
    hist_ret.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, hist_ret)
    return conb_df, hist_ret


def hist_ic_combine(factor_df, mret_df, factor_list, span, method='equal', half_life=1):
    """
    历史收IC加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算历史收益率均值
        method: 历史IC均值计算方法。'equal':算术平均；'half_life':半衰加权
        half_life: int, 半衰加权时的半衰期
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    
    # 计算历史因子收益率移动平均
    for fn in factor_list:
        if method == 'equal':
            ic_df[fn+'_ma'] = ic_df[fn].rolling(span).mean()
        elif method == 'half_life':
            ic_df[fn+'_ma'] = ic_df[fn].rolling(span).apply(lambda x: np.average(x, weights=list(reversed([0.5**(1.0*i/half_life) for i in range(span)]))))
        else:
            return 0
    
    ic_df = ic_df.dropna()
    ic_df = ic_df.drop(factor_list, axis=1)
    ic_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, ic_df)
    return conb_df, ic_df

def max_icir_combine(factor_df, mret_df, factor_list, span, method='sample', weight_limit=True):
    """
    最大化ICIR加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算IC均值和协方差矩阵
        method: 估计协方差矩阵的方法。'sample':直接用样本协方差矩阵；'shrunk':压缩估计
        weight_limit: bool, 是否约束权重为正
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    ic_df = ic_df.dropna()
    
    # 最大化ICIR
    m_ir_df = {}
    for i in range(span-1, len(ic_df)):
        dt = ic_df.ix[i, 'trade_date']
        ic_dt = ic_df.ix[i-span+1:i, factor_list]

        n = len(factor_list)
        # 求解最优化问题
        if method == 'sample':
            P = matrix(2*np.cov(ic_dt.T))
        elif method == 'shrunk':
            P = matrix(2*LedoitWolf().fit(ic_dt[factor_list].as_matrix()).covariance_)
        q = matrix([0.0]*n)
        G = matrix(-np.identity(n))
        h = matrix([0.0]*n)
        A = matrix(ic_dt.mean(), (1,n))
        b = matrix(1.0)
        if weight_limit:
            try:
                res = np.array(solvers.qp(P=P,q=q,G=G,h=h, A=A,b=b)['x'])
            except:
                res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])
        else:
            res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])

        m_ir_df[dt] = np.array(res).reshape(n)
    m_ir_df = pd.DataFrame(m_ir_df, index=factor_list).T.reset_index()
    if weight_limit:
        m_ir_df[factor_list] = np.where(m_ir_df[factor_list] < 0, 0, m_ir_df[factor_list])
    m_ir_df.loc[m_ir_df.sum(axis=1)==0, factor_list] = 1 
    m_ir_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, m_ir_df)
    return conb_df, m_ir_df

def max_ic_combine(factor_df, mret_df, factor_list, span, method='sample', weight_limit=True):
    """
    最大化IC加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算IC均值
        method: 估计协方差矩阵的方法。'sample':直接用样本协方差矩阵；'shrunk':压缩估计
        weight_limit: bool, 是否约束权重为正
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    for fn in factor_list:
        ic_df[fn] = ic_df[fn].rolling(span).mean()
    ic_df = ic_df.dropna()
    
    # 最大化IC
    m_ic_df = {}
    for dt in ic_df['trade_date']:
        ic_mean = ic_df.loc[ic_df['trade_date'] == dt, factor_list].values
        tmp_factor_df = factor_df.loc[factor_df['trade_date'] == dt, factor_list]

        n = len(factor_list)
        # 求解最优化问题
        if method == 'sample':
            P = matrix(2*np.cov(tmp_factor_df.T))
        elif method == 'shrunk':
            P = matrix(2*LedoitWolf().fit(tmp_factor_df.dropna().as_matrix()).covariance_)
        q = matrix([0.0]*n)
        G = matrix(-np.identity(n))
        h = matrix([0.0]*n)
        A = matrix(ic_mean, (1,n))
        b = matrix(1.0)
        if weight_limit:
            try:
                res = np.array(solvers.qp(P=P,q=q,G=G,h=h, A=A,b=b)['x'])
            except:
                res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])
        else:
            res = np.array(solvers.qp(P=P,q=q,A=A,b=b)['x'])

        m_ic_df[dt] = np.array(res).reshape(n)
    m_ic_df = pd.DataFrame(m_ic_df, index=factor_list).T.reset_index()
    if weight_limit:
        m_ic_df[factor_list] = np.where(m_ic_df[factor_list] < 0, 0, m_ic_df[factor_list])
    m_ic_df.loc[m_ic_df.sum(axis=1)==0, factor_list] = 1 
    m_ic_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, m_ic_df)
    return conb_df, m_ic_df


import pandas as pd
from sklearn.model_selection import KFold
from ultron.optimize.model.modelbase import load_module


def model_fitness(features, model_name, X, Y, params, mode, n_splits):
    kf = KFold(n_splits=n_splits, shuffle=False)
    res = []
    for train_index, test_index in kf.split(X):
        x_train = X.iloc[train_index]
        y_train = Y.iloc[train_index].values
        x_test = X.iloc[test_index]
        y_test = Y.iloc[test_index].values
        model = load_module(model_name)(features=features, **params)
        model.fit(x_train, y_train)
        res.append({
            'score': model.score(x_test, y_test),
            'accuracy': model.accuracy(x_test, y_test),
            'precision': model.precision(x_test, y_test),
            'recall': model.recall(x_test, y_test)
        })
    result = pd.DataFrame(res)
    mode_name = mode if mode in result.columns else 'accuracy'
    return 0 if result.empty else result[mode_name].mean()


from .stats import (aggregate_returns, annual_return, annual_volatility, cagr,
                    calmar_ratio, cum_returns, cum_returns_final,
                    downside_risk, excess_sharpe, max_drawdown, omega_ratio,
                    roll_annual_volatility, roll_max_drawdown,
                    roll_sharpe_ratio, roll_sortino_ratio, sharpe_ratio,
                    simple_returns, sortino_ratio, tail_ratio, _adjust_returns)

from .periods import (DAILY, WEEKLY, MONTHLY, QUARTERLY, YEARLY)

import numpy as np


def turnover(data,
             factor_name,
             key=['trade_date', 'code'],
             groupby_key='trade_date'):

    def to_weights(group):
        demeaned_vals = group - group.mean()
        return demeaned_vals / demeaned_vals.abs().sum()

    weights = data.set_index(key)[[
        factor_name
    ]].groupby(level=groupby_key).apply(to_weights)
    turnover_se = weights.unstack().diff().abs().sum(axis=1)
    return turnover_se.mean()


def fitness(returns_se, turnover=None, horizon=1):
    ir = returns_se.mean() / returns_se.std()
    sharpe = np.sqrt(252 / horizon) * ir
    returns = returns_se.sum() * 252 / horizon / len(returns_se)
    fitness = sharpe * np.sqrt(abs(returns) / turnover)
    return fitness


#def fitnesss(rets, turnover):
#    returns = factor_ret.sum()*252/horizon/len(factor_ret)
#    fitness = sharpe * np.sqrt(abs(returns)/turnover)



# -*- coding: utf-8 -*-
import unittest

import numpy as np
import pandas as pd

from ultron.factor.data.winsorize import NormalWinsorizer
from ultron.factor.data.winsorize import winsorize_normal


class Winsorize(unittest.TestCase):

    def setUp(self):
        self.x = np.random.randn(3000,10)
        self.groups = np.random.randint(10,30,size=3000)
        self.num_stds = 2
    
    def test_winsorize_normal(self):
        calc_winsorized = winsorize_normal(self.x, self.num_stds)

        std_values = self.x.std(axis=0, ddof=1)
        mean_value = self.x.mean(axis=0)
        
        lower_bound = mean_value - self.num_stds * std_values
        upper_bound = mean_value + self.num_stds * std_values

        for i in range(np.size(calc_winsorized, 1)):
            col_data = self.x[:,i]
            col_data[col_data > upper_bound[i]] = upper_bound[i]
            col_data[col_data < lower_bound[i]] = lower_bound[i]

            calculated_col = calc_winsorized[:, i]
            np.testing.assert_array_almost_equal(col_data, calculated_col)

    def test_winsorize_normal_with_group(self):
        cal_winsorized = winsorize_normal(self.x, self.num_stds, groups=self.groups)

        def impl(x):
            std_values = x.std(axis=0, ddof=1)
            mean_value = x.mean(axis=0)

            lower_bound = mean_value - self.num_stds * std_values
            upper_bound = mean_value + self.num_stds * std_values

            res = np.where(x > upper_bound, upper_bound, x)
            res = np.where(res < lower_bound, lower_bound, res)
            return res

        exp_winsorized = pd.DataFrame(self.x).groupby(self.groups).transform(impl).values
        np.testing.assert_array_almost_equal(cal_winsorized, exp_winsorized)

    def test_normal_winsorizer(self):
        s = NormalWinsorizer(num_stds=self.num_stds)
        s.fit(self.x)
        calc_winsorized1 = s.transform(self.x)
        calc_winsorized2 = s(self.x)

        std_values = self.x.std(axis=0, ddof=1)
        mean_value = self.x.mean(axis=0)

        lower_bound = mean_value - self.num_stds * std_values
        upper_bound = mean_value + self.num_stds * std_values

        for i in range(np.size(calc_winsorized1, 1)):
            col_data = self.x[:, i]
            col_data[col_data > upper_bound[i]] = upper_bound[i]
            col_data[col_data < lower_bound[i]] = lower_bound[i]

            calculated_col = calc_winsorized1[:, i]
            np.testing.assert_array_almost_equal(col_data, calculated_col)
            calculated_col = calc_winsorized2[:, i]
            np.testing.assert_array_almost_equal(col_data, calculated_col)


# -*- coding: utf-8 -*-

cpdef int pyFinAssert(condition, exception, str msg=*) except -1

cpdef int pyEnsureRaise(exception, str msg=*) except -1

cpdef int pyFinWarning(condition, warn_type, str msg=*)

cpdef bint isClose(double a, double b=*, double rel_tol=*, double abs_tol=*) nogil


# -*- coding: utf-8 -*-

import numpy as np
import time, datetime, pickle, itertools, os
from ultron.factor.genetic.geneticist.genetic import Gentic
from .operators import operators_sets
from ....utilities.mlog import MLog

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min


class Engine(object):

    def __init__(
            self,
            population_size=2000,
            generations=MAX_INT,
            tournament_size=20,
            stopping_criteria=0.0,
            factor_sets=None,
            init_depth=(5, 6),
            init_method='full',
            operators_set=operators_sets,
            n_jobs=1,
            p_crossover=0.9,
            p_subtree_mutation=0.01,
            p_hoist_mutation=0.01,
            p_point_mutation=0.01,
            p_point_replace=0.05,
            greater_is_better=True,  #True 倒序， False 正序
            verbose=1,
            is_save=1,
            rootid=0,
            session=0,
            standard_score=2,  # None代表 根据tournament_size保留种群  standard_score保留种群
            out_dir='result',
            backup_cycle=0,  # 后备数据周期，主要用于在时间序列上的问题
            convergence=None,  # 收敛值，若为None，则不需要收敛值。
            low_memory=False,
            fitness=None,
            random_state=None,
            custom_params=None,
            save_model=None):
        MLog().config(name='Gentic')
        self._population_size = population_size
        self._generations = MAX_INT if generations == 0 else generations
        self._tournament_size = tournament_size
        self._stopping_criteria = stopping_criteria
        self._factor_sets = factor_sets
        self._init_depth = init_depth
        self._init_method = init_method
        self._operators_set = operators_set
        self._function_set = [op.name for op in self._operators_set]
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._random_state = random_state
        self._greater_is_better = greater_is_better
        self._standard_score = standard_score
        self._fitness = fitness
        self._n_jobs = n_jobs
        self._backup_cycle = backup_cycle
        self._custom_params = custom_params
        self._low_memory = low_memory
        self._verbose = verbose
        self._is_save = is_save
        self._out_dir = out_dir
        self._convergence = convergence
        self._rootid = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if rootid == 0 else rootid
        self._session = int(
            time.time() * 1000000 +
            datetime.datetime.now().microsecond) if session == 0 else session
        self._save_model = self.save_model if save_model is None else save_model

    def run_gentic(self, total_data):
        gentic = Gentic(
            population_size=self._population_size,
            generations=self._generations,
            tournament_size=self._tournament_size,
            stopping_criteria=self._stopping_criteria,
            factor_sets=self._factor_sets,
            init_depth=self._init_depth,
            init_method=self._init_method,
            operators_set=self._operators_set,
            n_jobs=self._n_jobs,
            p_crossover=self._p_crossover,
            p_subtree_mutation=self._p_subtree_mutation,
            p_hoist_mutation=self._p_hoist_mutation,
            p_point_mutation=self._p_point_mutation,
            p_point_replace=self._p_point_replace,
            greater_is_better=self._greater_is_better,
            verbose=self._verbose,
            is_save=self._is_save,
            rootid=self._rootid,
            standard_score=self.
            _standard_score,  # None代表 根据tournament_size保留种群  standard_score保留种群
            out_dir=self._out_dir,
            backup_cycle=self._backup_cycle,  # 后备数据周期，主要用于在时间序列上的问题
            convergence=self._convergence,  # 收敛值，若为None，则不需要收敛值。
            low_memory=self._low_memory,
            fitness=self._fitness,
            session=self._session,
            random_state=self._random_state,
            custom_params=self._custom_params,
            save_model=self._save_model)

        gentic.train(total_data=total_data)
        result = gentic._run_details
        raw_fitness = 0 if len(
            result['best_programs']) == 0 else result['best_fitness'][-1]
        del gentic
        return raw_fitness

    def train(self, total_data):
        raw_fitness = 0
        while raw_fitness < self._stopping_criteria:
            raw_fitness = self.run_gentic(total_data)


import pandas as pd

def industry_fill(factor_data, factor_sets, index_key=['trade_date','code'],
                  group_key=['trade_date','industry_code'], fill_type='mean'):
    mean_values = factors_data.groupby(group_key).mean()[factor_columns].fillna(0)
    standard_data = factors_data.merge(mean_values.reset_index(),on=group_key, 
                                       how='left',suffixes=('', '_w')).set_index(
                                       index_key)
    mean_columns = [f + '_w'for f in factor_sets]
    factor_res = []
    for f in factor_columns:
        factor_data = standard_data[f].mask(pd.isnull(standard_data[f]),standard_data[f + '_w'])
        factor_res.append(factor_data)
    return pd.concat(factor_res, axis=1)
    
def data_processing(factor_name, total_data, is_ind_neutralize, custom_risk_styles,
                    pre_process=1, post_process=1, is_nansorize=1):
    alpha_res = []
    grouped = total_data.dropna(subset=[factor_name]).groupby(['trade_date']) if is_nansorize else \
                total_data.fillna(0).groupby(['trade_date'])
    ndiff_columns = [col for col in total_data.columns if col not in [factor_name]]
    ind_styles = industry_styles if is_ind_neutralize else []
    custom_risk_styles = [i for i in custom_risk_styles.split(',') if len(i) > 2] if len(custom_risk_styles) > 0 else []
    neutralized_styles = ind_styles + custom_risk_styles
    for k, g in grouped:
        if len(g) < 2:
            continue
        new_factors = factor_processing(g[[factor_name]].values,
                 pre_process=[winsorize_normal,standardize] if pre_process else None,
                 risk_factors=g[neutralized_styles].values.astype(float) if len(neutralized_styles) > 0 else None,
                 post_process=[winsorize_normal,standardize] if post_process else None)
        f = pd.DataFrame(new_factors, columns=[factor_name])
        for k in ndiff_columns:
            f[k] = g[k].values
        alpha_res.append(f)
    return pd.concat(alpha_res) if alpha_res else pd.DataFrame(columns=['trade_date',factor_name])


# -*- coding: utf-8 -*-

import warnings
from libc.math cimport fabs
from libc.math cimport fmax


cpdef int pyFinAssert(condition, exception, str msg="") except -1:
    if not condition:
        raise exception(msg)
    return 0


cpdef int pyEnsureRaise(exception, str msg="") except -1:
    raise exception(msg)


cpdef int pyFinWarning(condition, warn_type, str msg=""):
    if not condition:
        warnings.warn(msg, warn_type)
    return 0

cpdef bint isClose(double a, double b=0., double rel_tol=1e-09, double abs_tol=1e-12) nogil:
    return fabs(a-b) <= fmax(rel_tol * fmax(fabs(a), fabs(b)), abs_tol)



# -*- encoding:utf-8 -*-
"""
    统计相关工具模块
"""
import numpy as np
import functools
import scipy.stats as scs
from collections import namedtuple, Iterable


# noinspection PyClassHasNoInit
class MomentsTuple(
        namedtuple(
            'MomentsTuple',
            ('count', 'max', 'min', 'mean', 'std', 'skewness', 'kurtosis'))):
    __slots__ = ()

    def __repr__(self):
        return "count:{}\nmax:{}\nmin:{}\nmean:{}\nstd:{}\nskewness:{}\nkurtosis:{}".format(
            self.count, self.max, self.min, self.mean, self.std, self.skewness,
            self.kurtosis)


def arr_to_numpy(func):
    """
        函数装饰器：定参数装饰器，非通用，通用转换使用UltronDTUtil中的装饰器
        将被装饰函数中的arr序列转换为np.array
    """

    @functools.wraps(func)
    def wrapper(arr, *arg, **kwargs):
        # TODO Iterable和six.string_types的判断抽出来放在一个模块，做为Iterable的判断来使用
        if not isinstance(arr, Iterable) or isinstance(arr, six.string_types):
            # arr必须是可以迭代的对象
            raise TypeError('arr not isinstance of Iterable')

        if not isinstance(arr, np.ndarray):
            if isinstance(arr, pd.DataFrame) or isinstance(arr, pd.Series):
                # 如果是pandas直接拿values
                arr = arr.values
            elif isinstance(arr, dict):
                # 针对dict转换np.array
                arr = np.array(list(arr.values())).T
            else:
                arr = np.array(arr)
        return func(arr, *arg, **kwargs)

    return wrapper


@arr_to_numpy
def stats_namedtuple(arr):
    """
    通过序列构造arr的基础统计信息dict, 被arr_to_numpy装饰，统一输出，且这样使用arr.max(), arr.min()等不需要axis参数区别
    与stats_dict的区别只是返回namedtuple对象
    :param arr: pd.DataFrame or pd.Series or Iterable
    :return: MomentsTuple
                eg:
                    count:504
                    max:286.04
                    min:143.67
                    mean:228.48845238095237
                    std:25.538448192811927
                    skewness:-0.282635248604699
                    kurtosis:0.009313464006726946

    """
    count = arr.shape[0]
    if len(arr.shape) > 1 and arr.shape[1] > 1:
        count = arr.shape[0] * arr.shape[1]

    return MomentsTuple(count, arr.max(), arr.min(), arr.mean(), arr.std(),
                        scs.skew(arr), scs.kurtosis(arr))



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import copy
import ultron.factor.empyrical as empyrical
from ultron.strategy.executor.naive import NaiveExecutor
from ultron.optimize.riskmodel import FullRiskModel
from ultron.factor.covariance.cov_engine import CovEngine
from ultron.factor.experimental.portfolio import er_portfolio_analysis
from ultron.optimize.model.linearmodel import ConstLinearModel
from ultron.optimize.constraints import LinearConstraints, \
    create_box_bounds, BoundaryType
from ultron.tradingday import *
from ultron.utilities.logger import kd_logger


def create_params(**kwargs):
    turn_over_target = 1.0 if 'turn_over_target' not in kwargs else kwargs[
        'turn_over_target']
    target_vol = 0.1 if 'target_vol' not in kwargs else kwargs['target_vol']
    lbound = 0. if 'lbound' not in kwargs else kwargs['lbound']
    ubound = 0.04 if 'ubound' not in kwargs else kwargs['ubound']
    cov_windows = 20 if 'cov_windows' not in kwargs else kwargs['cov_windows']
    benchmark_lower = 1.001 if 'benchmark_lower' not in kwargs else kwargs[
        'benchmark_lower']
    benchmark_upper = 0.8 if 'benchmark_upper' not in kwargs else kwargs[
        'benchmark_upper']
    total_lower = -0.001 if 'total_lower' not in kwargs else kwargs[
        'total_lower']
    total_upper = 0.01 if 'total_upper' not in kwargs else kwargs['total_upper']
    effective_industry_lower = -0.20 if 'effective_industry_lower' not in kwargs else kwargs[
        'effective_industry_lower']
    effective_industry_upper = 0.20 if 'effective_industry_upper' not in kwargs else kwargs[
        'effective_industry_upper']
    invalid_industry_lower = -0.20 if 'invalid_industry_lower' not in kwargs else kwargs[
        'invalid_industry_lower']
    invalid_industry_upper = 0.20 if 'invalid_industry_upper' not in kwargs else kwargs[
        'invalid_industry_upper']
    weights_bandwidth = 0.1 if 'weights_bandwidth' not in kwargs else kwargs[
        'weights_bandwidth']
    cov_method = 'unshrunk' if 'cov_method' not in kwargs else kwargs[
        'cov_method']
    method = 'fmv' if 'method' not in kwargs else kwargs['method']
    effective_codes = [] if 'effective_codes' not in kwargs else kwargs[
        'effective_codes']
    invalid_codes = [] if 'invalid_codes' not in kwargs else kwargs[
        'invalid_codes']
    other_boundary = 'absolute' if 'other_boundary' not in kwargs else kwargs[
        'other_boundary']
    benchmark_boundary = 'absolute' if 'benchmark_boundary' not in kwargs else kwargs[
        'benchmark_boundary']
    total_boundary = 'absolute' if 'total_boundary' not in kwargs else kwargs[
        'total_boundary']
    effective_industry_boundary = 'absolute' if 'effective_industry_boundary' not in kwargs else kwargs[
        'effective_industry_boundary']
    invalid_industry_boundary = 'absolute' if 'invalid_industry_boundary' not in kwargs else kwargs[
        'invalid_industry_boundary']
    is_benchmark = 0 if 'is_benchmark' not in kwargs else kwargs['is_benchmark']

    params = {}
    params['industry'] = {}
    params['riskstyle'] = {}

    params['industry']['effective'] = effective_codes
    params['industry']['invalid'] = invalid_codes
    params['riskstyle']['effective'] = []
    params['riskstyle']['invalid'] = []

    params['setting_params'] = {}
    params['setting_params']['weights_bandwidth'] = weights_bandwidth
    params['setting_params']['method'] = method
    params['setting_params']['turn_over_target'] = turn_over_target
    params['setting_params']['target_vol'] = target_vol
    params['setting_params']['cov_windows'] = cov_windows
    params['setting_params']['cov_method'] = cov_method
    params['setting_params']['lbound'] = lbound
    params['setting_params']['ubound'] = ubound
    params['setting_params']['is_benchmark'] = is_benchmark

    params['setting_params']['benchmark'] = {}
    params['setting_params']['total'] = {}
    params['setting_params']['other'] = {}

    ###

    params['setting_params']['other']['boundary'] = other_boundary
    params['setting_params']['other']['lower'] = -0.3
    params['setting_params']['other']['upper'] = 0.3

    # benchmark 区间设置
    params['setting_params']['benchmark']['boundary'] = benchmark_boundary
    params['setting_params']['benchmark']['lower'] = benchmark_lower
    params['setting_params']['benchmark']['upper'] = benchmark_upper

    # total 区间设置   条件6
    params['setting_params']['total']['boundary'] = total_boundary
    params['setting_params']['total']['lower'] = total_lower
    params['setting_params']['total']['upper'] = total_upper

    ### 此处考虑行业择时
    params['setting_params']['effective_industry'] = {}
    params['setting_params']['invalid_industry'] = {}

    #### effective_industry 上限行业区间设置
    params['setting_params']['effective_industry'][
        'boundary'] = effective_industry_boundary
    params['setting_params']['effective_industry'][
        'lower'] = effective_industry_lower
    params['setting_params']['effective_industry'][
        'upper'] = effective_industry_upper

    #### invalid_industry 下限行业区间设置
    params['setting_params']['invalid_industry'][
        'boundary'] = invalid_industry_boundary
    params['setting_params']['invalid_industry'][
        'lower'] = invalid_industry_lower
    params['setting_params']['invalid_industry'][
        'upper'] = invalid_industry_upper
    return params


class RunningSetting(object):

    def __init__(self,
                 lbound=None,
                 ubound=None,
                 weights_bandwidth=None,
                 rebalance_method='risk_neutral',
                 bounds=None,
                 **kwargs):
        self.lbound = lbound
        self.ubound = ubound
        self.weights_bandwidth = weights_bandwidth
        self.rebalance_method = rebalance_method
        self.bounds = bounds
        self.more_opts = kwargs


class Strategy(object):

    def __init__(
        self,
        alpha_model,
        start_date=None,
        end_date=None,
        risk_model=None,
        total_data=None,
    ):
        self.start_date = start_date
        self.end_date = end_date
        self.alpha_model = alpha_model
        self.risk_model = risk_model
        self.total_data = total_data
        if self.total_data is not None:
            self.total_data['trade_date'] = pd.to_datetime(
                self.total_data['trade_date'])

    def prepare_backtest_data(self):
        pass

    def prepare_backtest_models(self, factor_name='factor'):
        models = {}
        total_data_groups = self.total_data.groupby('trade_date')
        alpha_model = ConstLinearModel(features=[factor_name],
                                       weights={factor_name: 1.0})
        for ref_date, _ in total_data_groups:
            models[ref_date] = alpha_model
        self.alpha_models = models
        kd_logger.info("alpha models training finished ...")

    def prepare_risk_model(self, name='unshrunk', window=20):
        models = {}
        returns_data = self.total_data[['trade_date', 'code', 'pev1_ret']]
        returns_data = returns_data.set_index(['trade_date', 'code']).unstack()
        returns_data_groups = returns_data.groupby('trade_date')
        for ref_date, _ in returns_data_groups:
            if ref_date < self.start_date:
                continue
            ref_begin_date = advanceDateByCalendar('china.sse', ref_date,
                                                   '-{0}b'.format(window))
            ref_end_date = advanceDateByCalendar('china.sse', ref_date, '-0b')
            rtb = returns_data.loc[ref_begin_date:ref_end_date].fillna(0)
            cov = CovEngine.calc_cov(name=name, ret_tb=rtb, window=window)
            model = FullRiskModel(cov)
            models[ref_date] = model
        self.risk_model = models

    def _build_setting(self, neutralized_styles, effective_industry,
                       invalid_industry, setting_params):
        _boundary = {
            'absolute': BoundaryType.ABSOLUTE,
            'relative': BoundaryType.RELATIVE
        }
        constraint_risk = neutralized_styles
        b_type = []
        l_val = []
        u_val = []
        total_risk_names = constraint_risk + ['benchmark', 'total']
        for name in total_risk_names:
            if name == 'benchmark':
                b_type.append(
                    _boundary[setting_params['benchmark']['boundary']])
                l_val.append(setting_params['benchmark']['lower'])
                u_val.append(setting_params['benchmark']['upper'])
            elif name == 'total':
                b_type.append(_boundary[setting_params['total']['boundary']])
                l_val.append(setting_params['total']['lower'])
                u_val.append(setting_params['total']['upper'])
            elif name in effective_industry:
                b_type.append(_boundary[setting_params['effective_industry']
                                        ['boundary']])
                l_val.append(setting_params['effective_industry']['lower'])
                u_val.append(setting_params['effective_industry']['upper'])
            elif name in invalid_industry:
                b_type.append(
                    _boundary[setting_params['invalid_industry']['boundary']])
                l_val.append(setting_params['invalid_industry']['lower'])
                u_val.append(setting_params['invalid_industry']['upper'])
            else:
                b_type.append(_boundary[setting_params['other']['boundary']])
            l_val.append(setting_params['other']['lower'])
            u_val.append(setting_params['other']['upper'])
        bounds = create_box_bounds(total_risk_names, b_type, l_val, u_val)
        return RunningSetting(
            bounds=bounds,
            lbound=setting_params['lbound'],
            ubound=setting_params['ubound'],
            weights_bandwidth=setting_params['weights_bandwidth'],
            rebalance_method=setting_params['method'],
            cov_winodws=setting_params['cov_windows'],
            cov_method=setting_params['cov_method'],
            target_vol=setting_params['target_vol'],
            is_benchmark=setting_params['is_benchmark'],
            turn_over_target=setting_params['turn_over_target'])

    @staticmethod
    def _create_lu_bounds(running_setting, codes, benchmark_w):

        codes = np.array(codes)

        if running_setting.weights_bandwidth:
            lbound = np.maximum(
                0., benchmark_w - running_setting.weights_bandwidth)
            ubound = running_setting.weights_bandwidth + benchmark_w

        lb = running_setting.lbound
        ub = running_setting.ubound

        if lb or ub:
            if not isinstance(lb, dict):
                lbound = np.ones_like(benchmark_w) * lb
            else:
                lbound = np.zeros_like(benchmark_w)
                for c in lb:
                    lbound[codes == c] = lb[c]

                if 'other' in lb:
                    for i, c in enumerate(codes):
                        if c not in lb:
                            lbound[i] = lb['other']
            if not isinstance(ub, dict):
                ubound = np.ones_like(benchmark_w) * ub
            else:
                ubound = np.ones_like(benchmark_w)
                for c in ub:
                    ubound[codes == c] = ub[c]

                if 'other' in ub:
                    for i, c in enumerate(codes):
                        if c not in ub:
                            ubound[i] = ub['other']
        return lbound, ubound

    def _calculate_pos(self,
                       running_setting,
                       er,
                       data,
                       constraints,
                       benchmark_w,
                       lbound,
                       ubound,
                       risk_model,
                       current_position,
                       is_benchmark=0):
        more_opts = running_setting.more_opts
        try:
            target_pos, _ = er_portfolio_analysis(
                er=er,
                industry=data.industry_code.values,
                dx_return=None,
                constraints=constraints,
                detail_analysis=False,
                benchmark=benchmark_w,
                method=running_setting.rebalance_method,
                lbound=lbound,
                ubound=ubound,
                current_position=current_position,
                target_vol=more_opts.get('target_vol'),
                turn_over_target=more_opts.get('turn_over_target'),
                risk_model=risk_model)
        except Exception:
            kd_logger.error('{0} rebalance error'.format(
                data.trade_date.values[0]))
            target_pos = current_position if not is_benchmark else benchmark_w
            target_pos = pd.DataFrame(target_pos, columns=['weight'])
            target_pos['industry'] = data.industry_code.values
            target_pos['er'] = er
        return target_pos

    ### 创建仓位持仓权重
    def create_positions(self, params):
        kd_logger.info("starting re-balance ...")
        total_data = copy.deepcopy(self.total_data)
        total_data = total_data if total_data is not None else copy.deepcopy(
            self.total_data)
        is_in_benchmark = (total_data.weight !=
                           0.).astype(float).values.reshape((-1, 1))
        total_data.loc[:, 'benchmark'] = is_in_benchmark
        total_data.loc[:, 'total'] = np.ones_like(is_in_benchmark)
        total_data_groups = self.total_data.groupby('trade_date')
        previous_pos = pd.DataFrame()
        positions = pd.DataFrame()

        running_setting = self._build_setting(
            neutralized_styles=params['industry']['effective'] +
            params['industry']['invalid'],
            effective_industry=params['industry']['effective'],
            invalid_industry=params['industry']['invalid'],
            setting_params=params['setting_params'])
        kd_logger.info("running setting finished ...")

        if self.alpha_model is None:
            self.prepare_backtest_models()

        if self.risk_model is None:
            self.prepare_risk_model(
                name=running_setting.more_opts['cov_method'],
                window=running_setting.more_opts['cov_winodws'])

        is_benchmark = running_setting.more_opts['is_benchmark']  \
                    if 'is_benchmark' in running_setting.more_opts else 0

        for ref_date, this_data in total_data_groups:
            if ref_date < self.start_date:
                continue
            new_model = self.alpha_models[ref_date]
            risk_model = self.risk_model[ref_date]
            codes = this_data.code.values.tolist()

            if previous_pos.empty:
                prev_date = advanceDateByCalendar('china.sse', ref_date, '-1b')
                current_position = self.total_data.set_index(
                    'trade_date').loc[prev_date].weight.values
            else:
                previous_pos.set_index('code', inplace=True)
                remained_pos = previous_pos.reindex(codes)
                remained_pos.fillna(0., inplace=True)
                current_position = remained_pos.weight.values
            benchmark_w = this_data.weight.values
            constraints = LinearConstraints(running_setting.bounds, this_data,
                                            benchmark_w)
            lbound, ubound = self._create_lu_bounds(running_setting, codes,
                                                    benchmark_w)
            this_data.fillna(0, inplace=True)
            ###是否标准化
            new_factors = this_data[new_model.features].values
            new_factors = pd.DataFrame(new_factors,
                                       columns=['factor'],
                                       index=codes)
            er = new_model.predict(new_factors).astype(float)

            kd_logger.info('{0} re-balance: {1} codes'.format(
                ref_date, len(er)))
            target_pos = self._calculate_pos(
                running_setting=running_setting,
                er=er,
                data=this_data,
                constraints=constraints,
                benchmark_w=benchmark_w,
                lbound=lbound,
                ubound=ubound,
                risk_model=risk_model.get_risk_profile(codes),
                current_position=current_position,
                is_benchmark=is_benchmark)
            target_pos['code'] = codes
            target_pos['trade_date'] = ref_date
            target_pos['benchmark'] = benchmark_w
            positions = positions.append(target_pos)
            previous_pos = target_pos
        return positions

    def matcher(self, ret_df):
        rets = []

        def match(returns, turnover=0, period=empyrical.DAILY):
            returns = returns.astype('float').dropna()
            annual_return = empyrical.annual_return(returns=returns,
                                                    period=period)
            annual_volatility = empyrical.annual_volatility(returns=returns,
                                                            period=period)
            cagr = empyrical.cagr(returns=returns, period=period)
            sharpe_ratio = empyrical.sharpe_ratio(returns=returns,
                                                  period=period)
            downside_risk = empyrical.downside_risk(returns=returns,
                                                    period=period)
            max_drawdown = empyrical.max_drawdown(returns=returns)
            results = {
                'annual_return': annual_return,
                'annual_volatility': annual_volatility,
                'cagr': cagr,
                'sharpe_ratio': sharpe_ratio,
                'downside_risk': downside_risk,
                'max_drawdown': max_drawdown,
                'calmar_ratio': -annual_return / max_drawdown
            }
            if turnover > 0.: results['turnover'] = turnover
            return results

        for r in ['returns', 'benchmark_returns', 'excess_return']:
            turnover = ret_df['turn_over'].mean() if r == 'returns' else 0
            mat = match(ret_df[r], turnover=turnover)
            mat['name'] = r
            rets.append(mat)
        return rets

    def backtest(self, positions, returns_data, rate):
        kd_logger.info("starting backting ...")
        executor = NaiveExecutor()
        total_data = positions.merge(
            returns_data[['trade_date', 'code', 'nxt1_ret']],
            on=['trade_date', 'code'])
        total_data_groups = total_data.groupby('trade_date')
        rets = []
        b_rets = []
        turn_overs = []
        leverags = []
        for ref_date, this_data in total_data_groups:
            turn_over, executed_pos = executor.execute(this_data)
            leverage = executed_pos.weight.abs().sum()
            ret = executed_pos.weight.values @ (
                np.exp(this_data.nxt1_ret.values) - 1.) - len(
                    this_data.code) * rate * turn_over
            b_ret = executed_pos.benchmark.values @ (
                np.exp(this_data.nxt1_ret.values) - 1.) - len(
                    this_data.code) * rate * turn_over
            rets.append(np.log(1. + ret))
            b_rets.append(np.log(1. + b_ret))
            executor.set_current(executed_pos)
            turn_overs.append(turn_over)
            leverags.append(leverage)
            kd_logger.info(
                '{0}: turn over {1}, returns {2}, benchmark returns {3}'.
                format(ref_date, round(turn_over, 4), round(ret, 4),
                       round(b_ret, 4)))
        trade_dates = positions.trade_date.unique()
        ret_df = pd.DataFrame(
            {
                'returns': rets,
                'turn_over': turn_overs,
                'leverage': leverags,
                'benchmark_returns': b_rets
            },
            index=trade_dates)
        ret_df['excess_return'] = ret_df[
            'returns'] - ret_df['benchmark_returns'] * ret_df['leverage']
        return ret_df

    def run(self, params, rate=0.0):
        positions = self.create_positions(params)
        returns_data = self.total_data[['trade_date', 'code', 'nxt1_ret']]
        ret_df = self.backtest(positions, returns_data, rate)
        match_df = self.matcher(ret_df)
        ret_df.loc[advanceDateByCalendar('china.sse', ret_df.index[-1],
                                         '1b')] = 0.
        ret_df = ret_df.shift(1)
        ret_df.iloc[0] = 0.
        return match_df, ret_df, positions
